# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import math

from msprobe.core.common.const import Const
from msprobe.visualization.graph.graph import Graph, BaseNode
from msprobe.visualization.graph.node_op import NodeOp
from msprobe.core.common.log import logger
from msprobe.visualization.utils import GraphConst, update_shared_dict, update_pbar_info
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.common.parallel_state import get_tp_pp_default_groups

MAX_INFO = 'The Max value merging method for '
MIN_INFO = 'The Min value merging method for '
MEAN_INFO = 'The Mean value merging method for '
NORM_INFO = 'The Norm value merging method for '


class GraphMerger:
    def __init__(self, build_graph_results, parallel_param, is_bench=False, pbar_info=None):
        self.strategy = self._select_strategy(build_graph_results, parallel_param, is_bench, pbar_info)

    @staticmethod
    def _select_strategy(results, param, is_bench, pbar_info):
        if param.tp == param.pp == param.rank_size == 1:
            return NoParallelMerger(results, param, is_bench, pbar_info)
        elif param.tp == param.rank_size:
            return TPMerger(results, param, is_bench, pbar_info)
        elif param.pp == param.rank_size:
            return PPMerger(results, param, is_bench, pbar_info) if param.vpp == 1 \
                else VPPMerger(results, param, is_bench, pbar_info)
        elif param.pp == 1:
            return TPMerger(results, param, is_bench, pbar_info)
        elif param.tp == 1:
            return PPMerger(results, param, is_bench, pbar_info) if param.vpp == 1 \
                else VPPMerger(results, param, is_bench, pbar_info)
        elif param.tp * param.pp == param.rank_size:
            return TPPPMerger(results, param, is_bench, pbar_info)
        else:
            return FullMerger(results, param, is_bench, pbar_info)

    def merge_graph(self):
        return self.strategy.merge_graphs()


class BaseGraphMerger:
    def __init__(self, build_graph_results, parallel_param, is_bench, pbar_info=None):
        self.unmerged_module = [Const.CLIP_GRAD, Const.OPTIMIZER]
        self.dtype_list = Const.TORCH_INT_DTYPE + Const.TORCH_FLOAT_DTYPE + [Const.FLOAT16, Const.FLOAT32,
                                                                             Const.BFLOAT16]
        self.build_graph_results = build_graph_results
        self.parallel_param = parallel_param
        self.is_bench = is_bench
        self.log_prefix = '[Bench]' if self.is_bench else '[NPU]'
        self.pbar_info = pbar_info
        self._add_all_nodes_rank()

    @staticmethod
    def sort_merged_api_collection(graph):
        def extract_rank(node):
            match = re.search(r'_Rank(\d+)', node.id)
            return int(match.group(1)) if match else None

        for sub_node in graph.root.subnodes:
            if sub_node.op == NodeOp.api_collection and sub_node.id.startswith(
                    GraphConst.APIS_BETWEEN_MODULES_ALL_RANKS):
                sub_node.subnodes = sorted(sub_node.subnodes, key=extract_rank)

    @staticmethod
    def _update_node_data_key(old_id, new_id, data_dict):
        new_dict = {}
        for key, value in data_dict.items():
            new_key = key.replace(old_id, new_id)
            if 'full_op_name' in value:
                value['full_op_name'] = value.get('full_op_name').replace(old_id, new_id)
            new_dict[new_key] = value
        return new_dict

    @staticmethod
    def _compare_value_same(main_value, other_value, has_uncertainty=False):
        if not isinstance(main_value, (int, float)) or not isinstance(other_value, (int, float)):
            return True
        # 没开启确定性计算，各rank的mean和norm有细微差异，如果相对误差在阈值内则认为是相同的
        if has_uncertainty:
            diff = abs(main_value - other_value)
            if math.isnan(diff):
                return math.isnan(main_value) and math.isnan(other_value)
            elif math.isinf(diff):
                return math.isinf(main_value) and math.isinf(other_value)
            else:
                return diff < GraphConst.UNCERTAINTY_THRESHOLD if main_value == 0 else \
                    abs(diff / main_value) < GraphConst.UNCERTAINTY_THRESHOLD
        else:
            return main_value == other_value

    def merge_graphs(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

    def merge_graph_api_collection(self, results: list):
        """
        graph合并时，将各rank的游离api集合合并为一个总的游离api集合
        example:
            rank0: Apis_Between_Modules.0                   rank1: Apis_Between_Modules.0
                   Module.module.Float16Module.forward.0           Module.module.Float16Module.forward.0
                   Apis_Between_Modules.1                          Apis_Between_Modules.1

            merged: Apis_Between_Modules_All_Ranks.0
                      |_ Apis_Between_Modules_Rank0.0
                      |_ Apis_Between_Modules_Rank1.0
                    Module.module.Float16Module.forward.0
                    Apis_Between_Modules_All_Ranks.1
                      |_ Apis_Between_Modules_Rank0.1
                      |_ Apis_Between_Modules_Rank1.1
        """
        main_graph_result = results[0]
        main_root_sub_nodes = main_graph_result.graph.root.subnodes
        new_main_root_sub_nodes = []
        for main_node in main_root_sub_nodes:
            # 如果游离api集合已合并为一个总的游离api集合，总的游离api集合之间还要再合并
            if main_node.id.startswith(GraphConst.APIS_BETWEEN_MODULES_ALL_RANKS):
                new_main_root_sub_nodes.append(main_node)
                for other_graph_result in results[1:]:
                    other_node = other_graph_result.graph.get_node(main_node.id)
                    if not other_node:
                        continue
                    for sub_node in other_node.subnodes:
                        sub_node.upnode = main_node
                        main_graph_result.graph.node_map[sub_node.id] = sub_node
                        for sub_sub_node in sub_node.subnodes:
                            main_graph_result.graph.node_map[sub_sub_node.id] = sub_sub_node
                    main_node.subnodes.extend(other_node.subnodes)
            # 游离api集合合并为一个总的游离api集合
            elif main_node.id.startswith(GraphConst.APIS_BETWEEN_MODULES):
                all_collection_node_id = main_graph_result.graph.add_node(NodeOp.api_collection,
                                                                          GraphConst.APIS_BETWEEN_MODULES_ALL_RANKS,
                                                                          id_accumulation=True)
                all_collection_node = main_graph_result.graph.get_node(all_collection_node_id)
                all_collection_node.upnode = main_graph_result.graph.root
                new_main_root_sub_nodes.append(all_collection_node)
                # Apis_Between_Modules.0 --> Apis_Between_Modules_Rank0.0
                origin_main_node_id = main_node.id
                main_node.id = GraphConst.APIS_BETWEEN_MODULES + f'_Rank{main_graph_result.rank}.' + \
                               main_node.id.split(Const.SEP)[-1]
                all_collection_node.subnodes = [main_node]
                main_node.upnode = all_collection_node
                main_graph_result.graph.node_map[main_node.id] = main_node
                del main_graph_result.graph.node_map[origin_main_node_id]
                for other_graph_result in results[1:]:
                    other_node = other_graph_result.graph.get_node(origin_main_node_id)
                    if not other_node:
                        continue
                    # Apis_Between_Modules.0 --> Apis_Between_Modules_Rank1.0
                    other_node.id = GraphConst.APIS_BETWEEN_MODULES + f'_Rank{other_graph_result.rank}.' + \
                                    other_node.id.split(Const.SEP)[-1]
                    main_graph_result.graph.node_map[other_node.id] = other_node
                    for sub_node in other_node.subnodes:
                        # api节点，在api名称上添加rank信息
                        old_id = sub_node.id
                        parts = sub_node.id.split(Const.SEP)
                        parts[1] += f'_rank{other_graph_result.rank}'
                        sub_node.id = Const.SEP.join(parts)
                        sub_node.input_data = self._update_node_data_key(old_id, sub_node.id, sub_node.input_data)
                        sub_node.output_data = self._update_node_data_key(old_id, sub_node.id, sub_node.output_data)
                        main_graph_result.graph.node_map[sub_node.id] = sub_node
                    all_collection_node.subnodes.append(other_node)
                    other_node.upnode = all_collection_node
            else:
                new_main_root_sub_nodes.append(main_node)
        main_graph_result.graph.root.subnodes = new_main_root_sub_nodes

    def split_graph_results_by_groups(self, groups):
        """
        基于pp或tp域，划分待合并的graph
        """
        rank_results_mapping = {result.rank: result for result in self.build_graph_results}
        return [[rank_results_mapping.get(rank) for rank in ranks] for ranks in groups]

    def compare_node_param_data(self, main_node, other_nodes, compare_data=True):
        """
        当前节点与若干其他节点比较输入输出参数的数据是否一致，如果发现有不一致的参数，将参数暂存于列表中
        :param main_node: 当前节点
        :param other_nodes: 其他节点列表
        :param compare_data: 是否进行数据比对，如果compare_data=False则直接认为数据不一致
        :return: 输入不一致的参数dict，输出不一致的参数dict，两个dict都为空列表代表两个节点的输入输出完全一致
        """
        if not other_nodes:
            return {}, {}
        data_types = {'input_data': {}, 'output_data': {}}
        for data_type, data_dict in data_types.items():
            main_data_dict = getattr(main_node, data_type)
            for key, main_param in main_data_dict.items():
                same_flag = compare_data
                if main_param.get(Const.DTYPE) not in self.dtype_list:
                    continue
                tp_need_merge_params = [main_param]
                for other_node in other_nodes:
                    param_key = key.replace(main_node.id, other_node.id) if main_node.id != other_node.id else key
                    other_param = getattr(other_node, data_type).get(param_key, {})
                    if other_param.get(Const.DTYPE) not in self.dtype_list:
                        break
                    tp_need_merge_params.append(other_param)
                    if compare_data and not self.compare_param_same(main_param, other_param, has_uncertainty=True):
                        same_flag = False
                if not same_flag:
                    data_dict[key.replace(main_node.id + Const.SEP, '')] = tp_need_merge_params
        return data_types.get('input_data'), data_types.get('output_data')

    def compare_param_same(self, main_param, other_param, has_uncertainty=False):
        if not self._compare_value_same(main_param.get(Const.MAX), other_param.get(Const.MAX)):
            return False
        if not self._compare_value_same(main_param.get(Const.MIN), other_param.get(Const.MIN)):
            return False
        if not self._compare_value_same(main_param.get(Const.MEAN), other_param.get(Const.MEAN), has_uncertainty):
            return False
        if not self._compare_value_same(main_param.get(Const.NORM), other_param.get(Const.NORM), has_uncertainty):
            return False
        return True

    def get_default_groups(self):
        """
        根据GPU总数、TP数、PP数初始化并行组

        return:
        tp_groups: 张量并行组列表，每个元素是一个包含组内rank的列表
        pp_groups: 流水线并行组列表，每个元素是一个包含组内rank的列表
        """
        tp_groups, pp_groups = get_tp_pp_default_groups(self.parallel_param.rank_size, self.parallel_param.tp,
                                                        self.parallel_param.pp, order=self.parallel_param.order)

        return tp_groups, pp_groups

    def _add_all_nodes_rank(self):
        for result in self.build_graph_results:
            for node in result.graph.node_map.values():
                node.rank = result.rank


class PPMerger(BaseGraphMerger):
    LAYERS_PATTERN = re.compile(r"(layers\.|layer\.)\d+(\.)")
    MARK_PATTERN = re.compile(r"%(\d+)%(\d+)$")
    MARK = '%'

    @staticmethod
    def _trace_p2p_mapping(p2p_mapping: dict):
        """
        将字典分组为独立的链，每个链都从未访问过的键开始，按照字典中的映射关系进行追踪
        p2p_mapping内容为p2p通信的send映射，追踪映射关系建立pp域
        example: p2p_mapping={0: 2, 1: 3, 2: 4, 3: 5, 4: 6, 5: 7, 6: 4, 7: 5}, return=[[0, 2, 4, 6], [1, 3, 5, 7]]
        """
        visited = set()
        result = []

        def collect_keys(start_key):
            """
            追踪从某一个键开始的所有“连续”键，直到无法再找到下一个键为止
            """
            current_key = start_key
            chain = []
            while current_key in p2p_mapping and current_key not in visited:
                chain.append(current_key)
                visited.add(current_key)
                current_key = p2p_mapping[current_key]
            return chain

        for key in p2p_mapping:
            if key not in visited:
                chain_result = collect_keys(key)
                if chain_result:
                    result.append(chain_result)
        return result

    @recursion_depth_decorator("msprobe.visualization.builder.graph_merger.PPMerger._merge_nodes", 1000)
    def _merge_nodes(self, main_graph, main_node, other_graphs):
        """
        其他rank graph中被pp切分的节点，需要合并到main graph
        """
        other_nodes = []
        for other_graph in other_graphs:
            other_node = other_graph.get_node(main_node.id)
            # 表明此节点只有main graph有
            if not other_node:
                other_nodes.clear()
                return
            other_nodes.append(other_node)
        if other_nodes:
            param_in, param_out = self.compare_node_param_data(main_node, other_nodes)
            # 各个rank都有的模块，且输入输出都不一致，且节点id符合正则，判定为被pp切分的模块，需要合并结构
            pp_merged_condition = param_in and param_out and self.LAYERS_PATTERN.search(main_node.id)
            # backward可能没有output，是否要pp合并从对应的forward节点判断
            if Const.SEP + Const.BACKWARD + Const.SEP in main_node.id:
                f_node = main_graph.node_map.get(
                    main_node.id.replace(Const.SEP + Const.BACKWARD + Const.SEP, Const.SEP + Const.FORWARD + Const.SEP))
                if f_node and hasattr(f_node, 'is_pp_merged'):
                    pp_merged_condition = True
            if pp_merged_condition:
                main_node.is_pp_merged = True
                main_up_node = main_node.upnode
                for other_node in other_nodes:
                    # pp切分中被切分的层在各rank的名称是一样的，这里给其他rank的同名层增加位置和rank标记
                    self._mark_node_id_position_rank(other_node, other_node.rank)
                    self._add_node_to_main_graph(main_graph, other_node)
                    # 其他rank被pp切分的模块节点添加到当前rank的graph
                    other_node.upnode = main_up_node
                    main_up_node.subnodes.append(other_node)
                # 已找到被pp切分的模块节点，不再递归其内部
                return
            # 各个rank都有的forward模块，且输入一致，输出不一致，判定为模块内部包含被pp切分的模块，此模块的输出要使用最后一个rank的输出
            elif not param_in and param_out and Const.SEP + Const.FORWARD + Const.SEP in main_node.id:
                main_node.output_data = other_nodes[-1].output_data
            # 各个rank都有的backward模块，且输出一致，输入不一致，判定为模块内部包含被pp切分的模块，此模块的输入要使用最后一个rank的输入
            elif param_in and not param_out and Const.SEP + Const.BACKWARD + Const.SEP in main_node.id:
                main_node.input_data = other_nodes[-1].input_data
            self._merge_other_unique_nodes(main_graph, main_node, other_nodes)
        for sub_node in main_node.subnodes:
            if sub_node.op == NodeOp.module:
                self._merge_nodes(main_graph, sub_node, other_graphs)

    def merge_graphs(self):
        results_groups = self.split_graph_results_by_groups(self.get_groups())
        results = []
        for result_groups in results_groups:
            self.merge_graph_api_collection(result_groups)
            results.extend(self.merge_pp_graphs(result_groups))
        return results

    def merge_pp_graphs(self, results):
        if self.pbar_info:
            update_shared_dict(self.pbar_info.current_stage_dict, self.pbar_info.task_id, 1, update_all=True)
        if not results or len(results) < 2:
            return results
        graphs = [x.graph for x in results]
        main_graph_result = results[0]
        total = len(main_graph_result.graph.root.subnodes)
        for i, main_node in enumerate(main_graph_result.graph.root.subnodes):
            if main_node.op == NodeOp.module and main_node.id not in self.unmerged_module:
                self._merge_nodes(main_graph_result.graph, main_node, graphs[1:])
                self._sort_nodes(main_graph_result.graph, main_node)
            if self.pbar_info:
                update_pbar_info(self.pbar_info, i + 1, total, update_all=True)
        return [main_graph_result]

    def get_groups(self):
        """
        在各rank寻找p2p通信节点，建立各rank之间p2p的映射关系
        """
        p2p_mapping = {}
        for result in self.build_graph_results:
            rank = result.rank
            pp_rank = None
            for node in result.graph.node_map.values():
                if not node.id.startswith(Const.DISTRIBUTED + Const.SEP):
                    continue
                if '.batch_isend_irecv.' in node.id:
                    for p2p_info in node.batch_p2p_info:
                        target_rank = p2p_info.get(GraphConst.PEER)
                        if target_rank is not None and target_rank != rank and p2p_info.get(GraphConst.OP) == 'isend':
                            pp_rank = target_rank
                            break
                elif '.send.' in node.id or '.isend.' in node.id:
                    # example: Distributed.isend.0.forward --> Distributed.isend.0.forward.input.dst
                    dst_kwarg = f'{node.id}{Const.SEP}{Const.INPUT}{Const.SEP}{GraphConst.DST}'
                    dst = node.input_data.get(dst_kwarg, {}).get('value')
                    if dst is not None:
                        pp_rank = dst
                        break
                if pp_rank is not None:
                    break
            if pp_rank is not None:
                p2p_mapping[rank] = pp_rank
        pp_groups = self._trace_p2p_mapping(p2p_mapping)
        if not pp_groups:
            logger.info('Unable to get pp groups based on Distributed Api (batch_isend_irecv, send, or isend), '
                        'generate pp groups using parallel param "rank_size", "tp" and "pp".')
            _, pp_groups = self.get_default_groups()
        elif len(pp_groups[0]) != self.parallel_param.pp:
            logger.warning(f'Based on Distributed Api (atch_isend_irecv, send, or isend), '
                           f'the resulting pp groups={pp_groups}, '
                           f'its length is not equal to the parallel param "pp"({self.parallel_param.pp}) you defined, '
                           f'generate pp groups using parallel param "rank_size", "tp" and "pp".')
            _, pp_groups = self.get_default_groups()
        logger.info(f'{self.log_prefix} All pp groups is {pp_groups}.')
        return pp_groups

    def _merge_other_unique_nodes(self, main_graph, main_node, other_nodes):
        """
        其他rank graph中other_node的子节点列表如果包含独有的节点，需要合并到main graph
        """
        lists = [main_node.subnodes]
        for other_node in other_nodes:
            lists.append(other_node.subnodes)
        dicts = [{node.id: node for node in lst} for lst in lists]
        unique_node_ids = {}
        # 计算每个集合的独有元素
        for i, current_dict in enumerate(dicts):
            other_ids = set()
            for j, other_dict in enumerate(dicts):
                if i != j:
                    # 更新并集，添加当前遍历到的集合的元素
                    other_ids.update(other_dict.keys())
            result = set(current_dict.keys()) - other_ids
            if i != 0 and result:
                # 计算当前集合与其他集合并集的差集，即独有元素，保持原始顺序
                unique_node_ids[i] = [node_id for node_id in current_dict if node_id in result]
        unique_nodes = []
        if unique_node_ids:
            for i, items in unique_node_ids.items():
                for item in items:
                    unique_nodes.append(dicts[i].get(item))
        if unique_nodes:
            for unique_node in unique_nodes:
                self._mark_node_id_position_rank(unique_node, unique_node.rank)
                self._add_node_to_main_graph(main_graph, unique_node)
                main_node.subnodes.append(unique_node)
                unique_node.upnode = main_node

    def _sort_nodes(self, main_graph, start_node):
        stack = [start_node]
        while stack:
            node = stack.pop()
            if self.MARK_PATTERN.search(node.id):
                is_forward = (Const.SEP + Const.FORWARD + Const.SEP in node.id or
                              Const.SEP + Const.FORWARD + self.MARK in node.id)
                new_sub_nodes1, new_sub_nodes2 = [], []
                for item in node.upnode.subnodes:
                    new_sub_nodes2.append(item) if self.MARK_PATTERN.search(item.id) else new_sub_nodes1.append(item)

                order = True if is_forward else False
                new_sub_nodes2.sort(key=lambda n: self._get_node_sort_rule(n, rank_ascending=order))
                new_sub_nodes = new_sub_nodes1 + new_sub_nodes2 if is_forward else new_sub_nodes2 + new_sub_nodes1

                index = -1
                node_iter = new_sub_nodes if is_forward else reversed(new_sub_nodes)
                for item in node_iter:
                    if self.LAYERS_PATTERN.search(item.id):
                        index += 1
                    if self.MARK_PATTERN.search(item.id):
                        item.pp_index = index

                for item in new_sub_nodes2:
                    self._update_node_id(main_graph, item)

                node.upnode.subnodes = new_sub_nodes

            stack.extend(node.subnodes)

    def _add_node_to_main_graph(self, main_graph: Graph, node: BaseNode):
        if node.id in main_graph.node_map:
            logger.warning(f'{node.id} is exist!')
        else:
            main_graph.node_map[node.id] = node
        for sub_node in node.subnodes:
            self._add_node_to_main_graph(main_graph, sub_node)

    def _get_node_sort_rule(self, node, rank_ascending=True):
        match = self.MARK_PATTERN.search(node.id)
        if match:
            # position代表当前节点在父节点中的位置序号
            position, rank = int(match.group(1)), int(match.group(2))
            if rank_ascending:
                return rank, position
            else:
                return -rank, position
        return (float('inf'), float('inf')) if rank_ascending else (-float('inf'), -float('inf'))

    def _mark_node_id_position_rank(self, node: BaseNode, rank):
        position = 0
        for index, item in enumerate(node.upnode.subnodes):
            if item.id == node.id:
                position = index
                break
        # 各rank重复节点添加所处层级位置排序信息position和rank号，用%分隔
        node.id = node.id + f'{self.MARK}{position}' + f'{self.MARK}{rank}'
        for sub_node in node.subnodes:
            self._mark_node_id_position_rank(sub_node, rank)

    def _update_node_id(self, graph, start_node: BaseNode, pp_index=""):
        stack = [(start_node, pp_index)]
        while stack:
            node, pp_index = stack.pop()
            # 修改节点id之前删除node_map的信息，修改完再添加回去
            if node.id not in graph.node_map:
                logger.warning(f'Update node id {node.id} fail!')
            else:
                del graph.node_map[node.id]
                old_id = self.MARK_PATTERN.sub("", node.id)
                if node.op == NodeOp.module:
                    # 被pp切分的模块节点，基于位置和rank信息修改模块名称计数信息
                    if self.LAYERS_PATTERN.search(node.id) and self.MARK_PATTERN.search(node.id):
                        if hasattr(node, 'pp_index'):
                            pp_index = str(node.pp_index)
                        node.id = self.LAYERS_PATTERN.sub(r"\g<1>" + pp_index + r"\g<2>", node.id)
                else:
                    # api节点，在api名称上添加rank信息
                    parts = node.id.split(Const.SEP)
                    parts[1] += f'_rank{node.id.split(PPMerger.MARK)[-1]}'
                    node.id = Const.SEP.join(parts)
                # 把之前添加的位置和rank信息删掉
                node.id = self.MARK_PATTERN.sub("", node.id)
                # node id更新了，那么data的key中包含node id也要更新
                node.input_data = self._update_node_data_key(old_id, node.id, node.input_data)
                node.output_data = self._update_node_data_key(old_id, node.id, node.output_data)
                graph.node_map[node.id] = node
            # 将子节点加入栈中
            for sub_node in node.subnodes:
                stack.append((sub_node, pp_index))


class TPMerger(BaseGraphMerger):
    RANK_PATTERN = re.compile(r"_rank(\d+)\.")
    OPERATION_TABLE = {
        Const.MAX: {
            'initial': lambda p: p.get(Const.MAX),
            'merge': lambda current, other: max(current, other.get(Const.MAX)),
            'finalize': lambda current, count: current,
            'formula': lambda key, values: f'{MAX_INFO}{key} is: max({", ".join(map(str, values))})'
        },
        Const.MIN: {
            'initial': lambda p: p.get(Const.MIN),
            'merge': lambda current, other: min(current, other.get(Const.MIN)),
            'finalize': lambda current, count: current,
            'formula': lambda key, values: f'{MIN_INFO}{key} is: min({", ".join(map(str, values))})'
        },
        Const.MEAN: {
            'initial': lambda p: p.get(Const.MEAN),
            'merge': lambda current, other: current + other.get(Const.MEAN),
            'finalize': lambda current, count: current / count,
            'formula': lambda key, values: f'{MEAN_INFO}{key} is: ({" + ".join(map(str, values))}) / {len(values)}'
        },
        Const.NORM: {
            'initial': lambda p: pow(p.get(Const.NORM), 2.0),
            'merge': lambda current, other: current + pow(other.get(Const.NORM), 2.0),
            'finalize': lambda current, count: pow(current, 1 / 2.0),
            'formula': lambda key, values: f'{NORM_INFO}{key} is: ({" + ".join([f"{v} ** 2" for v in values])}) ** 0.5'
        }
    }
    TP_MERGED_INFO = f'This data is the merged data after tensor parallelism(TP), and the data is merged from rank '

    @staticmethod
    def _merge_params(tp_need_merge_param: dict):
        """
        合并tp切分的各rank参数统计值
        tp_need_merge_param: {input.0: [{"Max": 0, "Min": 0, ...}, {"Max": 0.1, "Min": 0, ...}, ...]}
        return: 计算详情
        """
        merge_info = []
        for key, param_list in tp_need_merge_param.items():
            if len(param_list) < 2:
                continue
            main_param = param_list[0]

            for stat, ops in TPMerger.OPERATION_TABLE.items():
                current_value = ops['initial'](main_param)
                value_list = [current_value if stat != Const.NORM else main_param.get(Const.NORM)]

                for other_param in param_list[1:]:
                    current_value = ops['merge'](current_value, other_param)
                    value_list.append(other_param.get(stat) if stat != Const.NORM else other_param.get(Const.NORM))

                final_value = ops['finalize'](current_value, len(param_list))
                main_param[stat] = final_value
                formula_base = f'{ops["formula"](key, value_list)}' + f' = {final_value}'

                merge_info.append(formula_base)

        return merge_info

    @staticmethod
    def _get_need_merge_node(main_node, other_graphs, tp_merge_mapping):
        """
        获取需要TP合并的节点列表
        如果是TP+PP的混合并行，此时数据已经被PP合并过，一些node_id被标记上rank信息，此时需要基于rank映射才能获取到需要TP合并的节点列表，例如：
        main_node = Torch.matmul_rank4.32.forward  other_node = Torch.matmul_rank5.32.forward
        需要建立4->5的映射，才能基于Torch.matmul_rank4.32.forward找到Torch.matmul_rank5.32.forward
        """
        other_nodes = []
        match = TPMerger.RANK_PATTERN.search(main_node.id)
        # 节点名称被标记rank信息，且提供了映射
        if match and tp_merge_mapping:
            rank = int(match.group(1))
            tp_mapping_ranks = tp_merge_mapping.get(rank)
            if not tp_mapping_ranks:
                return other_nodes
            if len(tp_mapping_ranks) != len(other_graphs):
                return other_nodes
            for i, graph in enumerate(other_graphs):
                # 基于映射得到目标rank，替换node_id当前rank信息后去目标graph取node
                tp_mapping_id = TPMerger.RANK_PATTERN.sub(f"_rank{tp_mapping_ranks[i]}.", main_node.id)
                other_node = graph.node_map.get(tp_mapping_id)
                if not other_node or main_node.get_ancestors() != other_node.get_ancestors():
                    other_nodes.clear()
                    break
                other_nodes.append(other_node)
        else:
            for graph in other_graphs:
                other_node = graph.node_map.get(main_node.id)
                if not other_node or main_node.get_ancestors() != other_node.get_ancestors():
                    other_nodes.clear()
                    break
                other_nodes.append(other_node)

        return other_nodes

    @staticmethod
    def _slice_list_at_id(node_list, target_id1, target_id2):
        start_index, end_index = -1, -1
        for index, node in enumerate(node_list):
            if target_id1 in node.id:
                start_index = index
            elif target_id2 in node.id:
                end_index = index
        return [] if start_index == -1 or end_index == -1 else node_list[start_index:end_index + 1]

    def merge_graphs(self):
        results_groups = self.split_graph_results_by_groups(self.get_groups())
        results = []
        for result_groups in results_groups:
            self.merge_graph_api_collection(result_groups)
            results.extend(self.merge_tp_graphs(result_groups))
        return results

    def merge_tp_graphs(self, results, tp_merge_mapping=None):
        if self.pbar_info:
            update_shared_dict(self.pbar_info.current_stage_dict, self.pbar_info.task_id, 1, update_all=True)
        if not results or len(results) < 2:
            return results
        graphs = [x.graph for x in results]
        main_graph_result = results[0]
        total = len(main_graph_result.graph.node_map.values())
        for i, main_node in enumerate(main_graph_result.graph.node_map.values()):
            should_continue = (
                    not main_node.upnode or main_node.upnode.op != NodeOp.module or
                    main_node.upnode.id in self.unmerged_module or main_node.id.startswith(Const.DISTRIBUTED) or
                    main_node.parallel_merge_info != [])
            if should_continue:
                continue
            self._handle_tp_matmul_reduce(main_node, graphs[1:], tp_merge_mapping)
            other_nodes = self._get_need_merge_node(main_node, graphs[1:], tp_merge_mapping)
            tp_need_merge_param_in, tp_need_merge_param_out = self.compare_node_param_data(main_node, other_nodes)
            if tp_need_merge_param_in or tp_need_merge_param_out:
                ranks = [main_node.rank]
                for other_node in other_nodes:
                    ranks.append(other_node.rank)
                main_node.parallel_merge_info.append(f'{self.TP_MERGED_INFO}{ranks}.')
                merge_info_in = self._merge_params(tp_need_merge_param_in)
                merge_info_out = self._merge_params(tp_need_merge_param_out)
                main_node.parallel_merge_info.extend(merge_info_in + merge_info_out)
            if self.pbar_info:
                update_pbar_info(self.pbar_info, i + 1, total, update_all=True)
        for main_node in main_graph_result.graph.node_map.values():
            self._merge_tp_megatron_column_row_parallel(main_node, graphs[1:], tp_merge_mapping)
        return [main_graph_result]

    def get_groups(self):
        tp_groups = []
        for result in self.build_graph_results:
            for node in result.graph.node_map.values():
                if any(op in node.id for op in GraphConst.REDUCE_OPERATIONS):
                    group_ranks = node.input_data.get(f'{node.id}.input.group', {}).get('group_ranks')
                    if group_ranks and group_ranks not in tp_groups:
                        tp_groups.append(group_ranks)
                    break
        if not tp_groups:
            logger.info('Unable to get tp groups based on Distributed Api (reduce_scatter or all_reduce), '
                        'generate tp groups using parallel param "rank_size", "tp" and "pp".')
            tp_groups, _ = self.get_default_groups()
        elif len(tp_groups[0]) != self.parallel_param.tp:
            logger.warning(f'Based on Distributed Api (reduce_scatter or all_reduce), '
                           f'the resulting tp groups={tp_groups}, '
                           f'its length is not equal to the parallel param "tp"({self.parallel_param.tp}) you defined, '
                           f'generate tp groups using parallel param "rank_size", "tp" and "pp".')
            tp_groups, _ = self.get_default_groups()
        logger.info(f'{self.log_prefix} All tp groups is {tp_groups}.')
        return tp_groups

    def _handle_tp_matmul_reduce(self, node, other_graphs, tp_merge_mapping):
        """
        前向RowParallel和反向ColumnParallel层的matmul输出需要替换成matmul计算完成后all_reduce/reduce_scatter的输出
        """
        if node.op != NodeOp.module:
            return
        splits = node.id.split(Const.SEP)
        if len(splits) < 4:
            return
        is_forward_with_row_parallel = splits[-2] == Const.FORWARD and 'RowParallelLinear' in splits[-3]
        is_backward_with_column_parallel = splits[-2] == Const.BACKWARD and 'ColumnParallelLinear' in splits[-3]
        if not is_forward_with_row_parallel and not is_backward_with_column_parallel:
            return
        matmul_list = []
        reduce_list = []
        for sub_node in node.subnodes:
            if 'matmul' in sub_node.id:
                matmul_list.append(sub_node)
            if ('_reduce_scatter_base' in sub_node.id or 'reduce_scatter_tensor' in sub_node.id or
                    'all_reduce' in sub_node.id):
                reduce_list.append(sub_node)
        if not matmul_list or not reduce_list:
            return
        for matmul_node in matmul_list:
            if not matmul_node.output_data:
                continue
            # matmul的output0，将传递给all_reduce/reduce_scatter，作为all_reduce的input0，或作为reduce_scatter的input1
            matmul_node_output_param = list(matmul_node.output_data.values())[0]
            for reduce_node in reduce_list:
                if not reduce_node.output_data:
                    continue
                if 'all_reduce' in reduce_node.id:
                    if not reduce_node.input_data:
                        continue
                    reduce_node_input_param = list(reduce_node.input_data.values())[0]
                else:
                    if len(reduce_node.input_data) < 2:
                        continue
                    reduce_node_input_param = list(reduce_node.input_data.values())[1]
                if not self.compare_param_same(matmul_node_output_param, reduce_node_input_param):
                    continue
                # matmul的input统计值与其他rank的数据进行合并
                other_nodes = self._get_need_merge_node(matmul_node, other_graphs, tp_merge_mapping)
                tp_need_merge_param_in, _ = self.compare_node_param_data(matmul_node, other_nodes)
                if tp_need_merge_param_in:
                    ranks = [matmul_node.rank]
                    for other_node in other_nodes:
                        ranks.append(other_node.rank)
                    matmul_node.parallel_merge_info.append(f'{self.TP_MERGED_INFO}{ranks}.')
                    merge_info_in = self._merge_params(tp_need_merge_param_in)
                    matmul_node.parallel_merge_info.extend(merge_info_in)
                # matmul的output0替换为all_reduce/reduce_scatter的output0
                reduce_node_output_param = list(reduce_node.output_data.values())[0]
                keys = [Const.MAX, Const.MIN, Const.MEAN, Const.NORM]
                matmul_node_output_param.update({k: reduce_node_output_param.get(k) for k in keys})
                full_op_name = reduce_node_output_param.get('full_op_name')
                param_name = full_op_name if full_op_name else reduce_node.id
                matmul_node.parallel_merge_info.append(f'The output of this data is merged from {param_name}')
                reduce_list.remove(reduce_node)
                break

    def _merge_tp_megatron_column_row_parallel(self, node, other_graphs, tp_merge_mapping):
        if node.op != NodeOp.module or node.parallel_merge_info:
            return
        splits = node.id.split(Const.SEP)
        if len(splits) < 4:
            return
        is_forward_with_column_parallel = splits[-2] == Const.FORWARD and 'ColumnParallelLinear' in splits[-3]
        if not is_forward_with_column_parallel:
            return
        if not node.upnode:
            return
        # 获取[ColumnParallelLinear, RowParallelLinear]结构
        nodes = self._slice_list_at_id(node.upnode.subnodes, node.id, 'RowParallelLinear')
        if len(nodes) < 2:
            return
        stack = nodes[:]
        while stack:
            current_node = stack.pop()
            stack.extend(reversed(current_node.subnodes))

            if current_node.parallel_merge_info or current_node.id.startswith(Const.DISTRIBUTED):
                continue

            other_nodes = self._get_need_merge_node(current_node, other_graphs, tp_merge_mapping)
            param_in, param_out = self.compare_node_param_data(current_node, other_nodes, False)

            if param_in or param_out:
                ranks = [current_node.rank]
                for other_node in other_nodes:
                    ranks.append(other_node.rank)
                current_node.parallel_merge_info.append(f'{self.TP_MERGED_INFO}{ranks}.')
                # ColumnParallelLinear层的输入、其中的matmul输入不需要合并
                if current_node == nodes[0] or ('matmul' in current_node.id and current_node.upnode == nodes[0]):
                    param_in.pop('input.0', None)
                # RowParallelLinear层的输出、其中的matmul输出不需要合并, bias不需要合并
                elif current_node == nodes[-1] or ('matmul' in current_node.id and current_node.upnode == nodes[-1]):
                    param_out = {}
                    param_in.pop('parameters.bias', None)

                merge_info_in = self._merge_params(param_in)
                merge_info_out = self._merge_params(param_out)
                current_node.parallel_merge_info.extend(merge_info_in + merge_info_out)


class NoParallelMerger(BaseGraphMerger):
    def merge_graphs(self):
        self.merge_graph_api_collection(self.build_graph_results)
        return self.build_graph_results


class TPPPMerger(BaseGraphMerger):
    def merge_graphs(self):
        tp_merger = TPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info)
        pp_merger = PPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info) \
            if self.parallel_param.vpp == 1 else \
            VPPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info)
        pp_groups = pp_merger.get_groups()
        tp_groups = tp_merger.get_groups()
        # 进入TP+PP混合处理器，PP和TP必然大于1
        tp_merge_mapping = {}
        for tp_group in tp_groups[1:]:
            tp_merge_mapping[tp_group[0]] = tp_group[1:]
        self.merge_graph_api_collection(self.build_graph_results)
        # 先合并pp，需要知道pp域，在各自pp域中合并
        results_groups_pp = self.split_graph_results_by_groups(pp_groups)
        pp_results = []
        for results in results_groups_pp:
            pp_results.extend(pp_merger.merge_pp_graphs(results))
        # pp合并完成后，直接进行tp合并，最终得到一个graph
        tp_result = tp_merger.merge_tp_graphs(pp_results, tp_merge_mapping)
        self.sort_merged_api_collection(tp_result[0].graph)
        return tp_result


class FullMerger(BaseGraphMerger):
    def merge_graphs(self):
        tp_merger = TPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info)
        pp_merger = PPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info) \
            if self.parallel_param.vpp == 1 else \
            VPPMerger(self.build_graph_results, self.parallel_param, self.is_bench, self.pbar_info)
        pp_groups = pp_merger.get_groups()
        tp_groups = tp_merger.get_groups()
        tp_merge_mapping = {}
        if len(tp_groups) < 1:
            raise RuntimeError(f'Graph merged error, and tp_groups is {tp_groups}.')
        for tp_group in tp_groups[1:]:
            if len(tp_group) < 1:
                raise RuntimeError(f'Graph merged error, and tp_group is {tp_group}.')
            tp_merge_mapping[tp_group[0]] = tp_group[1:]
        # 先合并pp，需要知道pp域，在各自pp域中合并
        results_groups_pp = self.split_graph_results_by_groups(pp_groups)
        pp_results = {}
        for pp_result in results_groups_pp:
            self.merge_graph_api_collection(pp_result)
            pp_result = pp_merger.merge_pp_graphs(pp_result)[0]
            pp_results[pp_result.rank] = pp_result
        # pp合并完成后，基于tp域划分pp合并结果
        lists_to_be_tp_merged = []
        for tp_group in tp_groups:
            list_to_be_tp_merged = []
            for rank in tp_group:
                pp_result = pp_results.get(rank)
                if pp_result:
                    list_to_be_tp_merged.append(pp_result)
            if list_to_be_tp_merged:
                lists_to_be_tp_merged.append(list_to_be_tp_merged)
        tp_results = []
        for list_to_be_tp_merged in lists_to_be_tp_merged:
            self.merge_graph_api_collection(list_to_be_tp_merged)
            tp_merged_result = tp_merger.merge_tp_graphs(list_to_be_tp_merged, tp_merge_mapping)
            self.sort_merged_api_collection(tp_merged_result[0].graph)
            tp_results.extend(tp_merged_result)
        return tp_results


class VPPMerger(PPMerger):
    LAYERS_NUM_PATTERN = re.compile(r"(layers\.|layer\.)(\d+)(\.)")
    FORWARD_PATTERN = re.compile(r'\.forward\.\d+$')

    @staticmethod
    def _replace_vpp_id(s, vpp_id):
        parts = s.split(Const.SEP)
        if len(parts) < 2 or not parts[1].isdigit():
            return s
        parts[1] = str(vpp_id)
        return Const.SEP.join(parts)

    def merge_pp_graphs(self, results):
        if self.pbar_info:
            update_shared_dict(self.pbar_info.current_stage_dict, self.pbar_info.task_id, 1, update_all=True)
        if not results or len(results) < 2:
            return results
        graphs = [x.graph for x in results]
        main_graph_result = results[0]
        total = len(main_graph_result.graph.root.subnodes)
        for i, main_node in enumerate(main_graph_result.graph.root.subnodes):
            if main_node.op == NodeOp.module and main_node.id not in self.unmerged_module:
                self._merge_nodes(main_graph_result.graph, main_node, graphs[1:])
                self._sort_nodes(main_graph_result.graph, main_node)
            if self.pbar_info:
                update_pbar_info(self.pbar_info, i + 1, total, update_all=True)
        self._merge_vpp_data(main_graph_result.graph)
        self._merge_vpp_chunks(main_graph_result.graph)
        return [main_graph_result]

    def _merge_vpp_data(self, graph):
        """
        所有chunk的数据都合并到chunk0，前向chunk0的输出使用最后一个chunk的输出，反向chunk0的输入使用最后一个chunk的输入
        """
        module_list = []
        for node in reversed(graph.root.subnodes):
            parts = node.id.split(Const.SEP)
            if len(parts) < 2:
                continue
            if parts[1] in [GraphConst.VPP_CHUNK_0, str(self.parallel_param.vpp - 1)]:
                module_list.append(node)
        if not module_list:
            return
        stack = module_list[:]
        while stack:
            current_node = stack.pop()
            if hasattr(current_node, 'is_pp_merged') or hasattr(current_node,
                                                                'pp_index') or current_node.op != NodeOp.module:
                continue
            is_forward = self.FORWARD_PATTERN.search(current_node.id)
            stack.extend(reversed(current_node.subnodes))
            target_id = self._replace_vpp_id(current_node.id, self.parallel_param.vpp - 1)
            target_node = graph.node_map.get(target_id)
            if not target_node:
                continue
            if is_forward:
                current_node.output_data = self._update_node_data_key(target_node.id, current_node.id,
                                                                      target_node.output_data)
            else:
                current_node.input_data = self._update_node_data_key(target_node.id, current_node.id,
                                                                     target_node.input_data)

    def _merge_vpp_chunks(self, graph):
        """
        所有chunk都合并到chunk0，layers层搬到chunk0并重排序号
        """
        chunk_id_list = [i for i in range(1, self.parallel_param.vpp)]
        chunk_0_list = []
        for node in reversed(graph.root.subnodes):
            parts = node.id.split(Const.SEP)
            if len(parts) < 2:
                continue
            if parts[1] == GraphConst.VPP_CHUNK_0:
                chunk_0_list.append(node)
        if not chunk_0_list:
            return
        stack = chunk_0_list[:]
        layers_need_merge_dict = {}
        while stack:
            current_node = stack.pop()
            if hasattr(current_node, 'is_pp_merged') or hasattr(current_node, 'pp_index') \
                    and current_node.upnode.id not in layers_need_merge_dict:
                layers_need_merge_dict[current_node.upnode.id] = current_node.upnode
                continue
            stack.extend(reversed(current_node.subnodes))
        for node in layers_need_merge_dict.values():
            is_forward = self.FORWARD_PATTERN.search(node.id)
            for vpp_id in chunk_id_list:
                target_node = graph.node_map.get(self._replace_vpp_id(node.id, vpp_id))
                if not target_node:
                    continue
                # 其他chunk的layers都搬到chunk0，forward追加到后面，backward追加到前面
                if is_forward:
                    node.subnodes.extend(target_node.subnodes)
                else:
                    node.subnodes = target_node.subnodes + node.subnodes
                for sub_node in target_node.subnodes:
                    sub_node.upnode = node
                # 获取其他chunk的层级链路，删除所有父节点，不在前端展示已合并的其他chunk节点
                ancestors = target_node.get_ancestors()
                if len(ancestors) < 2:
                    continue
                for module_id in ancestors[1:]:
                    graph.node_map.pop(module_id, None)
                graph.root.subnodes = [node for node in graph.root.subnodes if node.id != ancestors[1]]
            # layers层重排序号
            self._sort_layers(node.subnodes, graph, is_forward)

    def _sort_layers(self, node_list, graph, is_forward):
        if not is_forward:
            node_list = list(reversed(node_list))
        index = -1
        for node in node_list:
            match = self.LAYERS_NUM_PATTERN.search(node.id)
            if match:
                index += 1
            parts = node.id.split(Const.SEP)
            # Module.0.xxx代表第一个chunk，不必重排序
            if len(parts) < 2 or parts[1] == GraphConst.VPP_CHUNK_0:
                continue
            # layers层修改chunk号和layers序号，非layers层修改chunk号
            new_node_id_prefix = ''
            if match:
                prefix, number, dot = match.groups()
                new_string = prefix + str(index) + dot
                start, end = match.span()
                new_node_id_prefix = node.id[:start] + new_string
                new_node_id_prefix = self._replace_vpp_id(new_node_id_prefix, GraphConst.VPP_CHUNK_0)
                new_node_id = new_node_id_prefix + node.id[end:]
            else:
                new_node_id = self._replace_vpp_id(node.id, GraphConst.VPP_CHUNK_0)
            graph.node_map.pop(node.id, None)
            node.input_data = self._update_node_data_key(node.id, new_node_id, node.input_data)
            node.output_data = self._update_node_data_key(node.id, new_node_id, node.output_data)
            node.id = new_node_id
            graph.node_map[new_node_id] = node
            stack = node.subnodes[:]
            while stack:
                current_node = stack.pop()
                if current_node.op != NodeOp.module:
                    continue
                stack.extend(reversed(current_node.subnodes))
                match = self.LAYERS_NUM_PATTERN.search(current_node.id)
                if match:
                    _, e = match.span()
                    new_current_node_id = new_node_id_prefix + current_node.id[e:]
                else:
                    new_current_node_id = self._replace_vpp_id(current_node.id, GraphConst.VPP_CHUNK_0)
                current_node.input_data = self._update_node_data_key(current_node.id, new_current_node_id,
                                                                     current_node.input_data)
                current_node.output_data = self._update_node_data_key(current_node.id, new_current_node_id,
                                                                      current_node.output_data)
                graph.node_map.pop(current_node.id, None)
                current_node.id = new_current_node_id
                graph.node_map[new_current_node_id] = current_node



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch

from msprobe.pytorch.dump.debugger.precision_debugger import PrecisionDebugger, module_dump, module_dump_end
from .common.utils import seed_all
from .torchair_dump import set_fx_dump_config, set_ge_dump_config

torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'
if torch_version_above_or_equal_2:
    from msprobe.pytorch.monitor.module_hook import TrainerMon

try:
    from msprobe.pytorch.aclgraph_dump import acl_save  # noqa: F401
except Exception:
    acl_save = None



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import re

from msprobe.core.common.log import logger
from msprobe.core.common.const import FileCheckConst
from msprobe.core.common.file_utils import check_file_or_directory_path, check_output_dir_path, check_file_type


def check_model_path_legality(path):
    check_file_or_directory_path(path, False, False, [".onnx", ".om"])
    return path


def check_input_data_path(path):
    if not isinstance(path, str):
        logger.error(f"input data path:{path} is illegal. Please check.")
        raise argparse.ArgumentTypeError
    if path == '':
        return path
    input_item_paths = path.split(',')
    for input_item_path in input_item_paths:
        file_type = check_file_type(input_item_path)
        if file_type != FileCheckConst.FILE:
            logger.error("The '--input_data' parameter only supports file paths, not folder paths. Please check.")
            raise argparse.ArgumentTypeError
        check_file_or_directory_path(input_item_path, False, False, [".npy", ".bin"])
    return path


def check_output_path_legality(path):
    if not path:
        return path
    check_output_dir_path(path)
    return path


def check_dict_kind_string(value):
    # just like "input_name1:1,224,224,3;input_name2:3,300"
    if not value:
        return value
    regex = re.compile(r"[^_A-Za-z0-9,;:/.-]")
    if regex.search(value):
        logger.error(f"string '{value}' is not a legal string")
        raise argparse.ArgumentTypeError
    return value


def check_rank_range_valid(rank):
    min_value = 0
    max_value = 255
    try:
        int_rank = int(rank)
    except Exception as e:
        logger.error("The rank is illegal. The value must be of type that can be cast to an int variable!")
        raise argparse.ArgumentTypeError from e
    if int_rank < min_value or int_rank > max_value:
        logger.error(f"rank:{rank} is invalid. valid value range is [{min_value}, {max_value}]")
        raise argparse.ArgumentTypeError
    return rank


def check_number_list(value):
    # just like "1241414,124141,124424"
    if not value:
        return value
    outsize_list = value.split(',')
    for outsize in outsize_list:
        regex = re.compile(r"[^0-9]")
        if regex.search(outsize):
            logger.error(f"output size \"{outsize}\" is not a legal string")
            raise argparse.ArgumentTypeError
    return value


def check_dym_range_string(value):
    if not value:
        return value
    dym_string = value
    regex = re.compile(r"[^_A-Za-z0-9,;:/.\-~]")
    if regex.search(dym_string):
        logger.error(f"dym range string \"{dym_string}\" is not a legal string")
        raise argparse.ArgumentTypeError
    return value


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        logger.error('Boolean value expected true, 1, false, 0 with case insensitive.')
        raise argparse.ArgumentTypeError



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import csv
import os
import re
from datetime import datetime, timezone
from typing import Any, Dict, List

from msprobe.core.common.const import Const
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.file_utils import make_dir
from msprobe.core.common.log import logger


class CSVWriterV2:
    """
    Minimal CSV writer for monitor_v2.

    - Each rank gets its own directory: <out_dir>/rank_<rank>/
    - Each monitor writes to <slug>_step<start>-<end>.csv
    - Row structure: step plus dynamic fields from rows (no ts/monitor/rank).
    """

    def __init__(self, out_dir: str = "monitor_v2_output", rank: int = 0, async_write: bool = False):
        self.out_dir = out_dir
        self.rank = rank
        self.async_write = async_write  # reserved, currently synchronous
        if self.async_write:
            logger.warning("[monitor_v2] async_write is not supported yet; falling back to sync writes.")

        self.rank_dir = os.path.join(self.out_dir, f"rank_{self.rank}")
        make_dir(self.rank_dir)
        self._known_fields: Dict[str, List[str]] = {}

    @staticmethod
    def close():
        return None

    @staticmethod
    def _stack_tensors(tensors: List[Any]) -> Any:
        if FmkAdp.fmk == Const.PT_FRAMEWORK:
            import torch

            return torch.stack(tensors, dim=0)
        from mindspore import ops

        return ops.stack(tensors, axis=0)

    def write_monitor_data(self, monitor_data: Dict[str, Any]):
        rows = monitor_data.get("rows") or []
        if not rows:
            return None
        rows = self._flatten_rows(rows)
        if not rows:
            return None

        monitor_name = monitor_data.get("monitor", "monitor")
        slug = self._safe_slug(monitor_data.get("slug", monitor_name))
        step = monitor_data.get("step", "unknown")
        interval = self._safe_int(monitor_data.get("step_count_per_record", 1), default=1)
        start_step = self._safe_int(monitor_data.get("start_step", 0), default=0)
        csv_path = self._resolve_csv_path(slug, step, interval, start_step)

        base_fields = ["vpp_stage", "step"]
        row_keys = list(dict.fromkeys(k for r in rows for k in r if k not in base_fields))
        fields = base_fields + row_keys

        file_exists = os.path.exists(csv_path)
        existing = self._known_fields.get(slug)
        if existing:
            if not file_exists:
                for k in fields:
                    if k not in existing:
                        existing.append(k)
            fields = existing
        else:
            self._known_fields[slug] = fields

        with open(csv_path, "a", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fields, extrasaction="ignore")
            if not file_exists:
                writer.writeheader()
            for r in rows:
                out = dict(r)
                out.setdefault("step", step)
                writer.writerow(out)

        return csv_path

    @staticmethod
    def _safe_int(value: Any, default: int = 0) -> int:
        try:
            return int(value)
        except (TypeError, ValueError):
            return default

    @staticmethod
    def _safe_slug(slug: Any) -> str:
        text = str(slug) if slug is not None else "monitor"
        text = text.replace(os.sep, "_")
        if os.altsep:
            text = text.replace(os.altsep, "_")
        text = text.replace("..", "_")
        text = re.sub(r"[^A-Za-z0-9_.-]+", "_", text).strip("._")
        return text or "monitor"

    def _resolve_csv_path(self, slug: str, step: Any, interval: int, start_step: int) -> str:
        try:
            step_val = int(step)
        except (TypeError, ValueError):
            return os.path.join(self.rank_dir, f"{slug}.csv")
        interval = max(interval, 1)
        relative = step_val - start_step
        if relative < 0:
            start = step_val
        else:
            start = start_step + (relative // interval) * interval
        end = start + interval - 1
        return os.path.join(self.rank_dir, f"{slug}_step{start}-{end}.csv")

    def _flatten_rows(self, rows: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        stat_ops = list(dict.fromkeys(op for row in rows for op in (row.get("stats") or {})))
        stacked_values = self._stack_stats(rows, stat_ops)
        default_values = [float("nan")] * len(rows)
        flattened = []
        for idx, row in enumerate(rows):
            base = {k: v for k, v in row.items() if k != "stats"}
            for op in stat_ops:
                values = stacked_values.get(op, default_values)
                base[op] = values[idx] if idx < len(values) else float("nan")
            flattened.append(base)
        return flattened

    def _stack_stats(self, rows: List[Dict[str, Any]], stat_ops: List[str]) -> Dict[str, List[float]]:
        stacked: Dict[str, List[float]] = {}
        total = len(rows)
        for op in stat_ops:
            tensors = [(row.get("stats") or {}).get(op) for row in rows]
            if not tensors or any(tensor is None for tensor in tensors):
                stacked[op] = [float("nan")] * total
                continue
            stacked_tensor = self._stack_tensors(tensors)
            if FmkAdp.fmk == Const.PT_FRAMEWORK:
                stacked_tensor = stacked_tensor.cpu()
            numpy_vals = FmkAdp.asnumpy(stacked_tensor).reshape(-1)
            stacked[op] = [float(value) for value in numpy_vals.tolist()]
        return stacked



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
MultiConvertProcess class. This class mainly involves the process function.
"""
import os
import multiprocessing
import psutil
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils

from msprobe.msaccucmp.cmp_utils.multi_process.progress import Progress
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class MultiConvertProcess:
    """
    The class for multi process for convert
    """
    MAX_MULTI = 4
    MULTI_THREAD_RESULT_COUNT = 2
    MULTI_THREAD_RETURN_CODE_INDEX = 0
    MULTI_THREAD_ERROR_FILE_INDEX = 1

    def __init__(self: any, process_func: any, input_path: list, output_path: str) -> None:
        self._process_func = process_func
        self._output_path = output_path
        self._input_path = input_path
        self._progress = None

    def process(self: any) -> int:
        """
        Process by multi process
        """
        return_code = CompareError.MSACCUCMP_NONE_ERROR

        # split big file and common file
        multi_process_file_list, big_file_list = self._split_big_file()

        if len(multi_process_file_list) > 0:
            ret = self._do_multi_process(multi_process_file_list)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                return_code = ret
        # big file do not multi process
        for big_file in big_file_list:
            ret, _ = self._process_func(big_file)
            self._handle_result_callback([ret, big_file])
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                return_code = ret
        if return_code != CompareError.MSACCUCMP_NONE_ERROR:
            error_file_path = os.path.join(self._output_path, ConstManager.CONVERT_FAILED_FILE_LIST_NAME)
            if os.path.exists(error_file_path):
                log.print_info_log('The list of files that failed to be converted has been written to "%r".'
                                   % error_file_path)

        return return_code

    def get_max_file_size(self: any) -> int:
        """
        Get max file size
        :return int
        """
        mem = psutil.virtual_memory()
        available = mem.available
        cpu_count = int((multiprocessing.cpu_count() + 1) / 2)
        if cpu_count != 0:
            return available / cpu_count / self.MAX_MULTI
        else:
            return 0

    def _handle_result_callback(self: any, result: list) -> None:
        self._progress.update_progress()
        self._progress.update_and_print_progress()
        if len(result) == self.MULTI_THREAD_RESULT_COUNT:
            cur_ret = result[self.MULTI_THREAD_RETURN_CODE_INDEX]
            if cur_ret != CompareError.MSACCUCMP_NONE_ERROR:
                error_file_path = os.path.join(self._output_path, ConstManager.CONVERT_FAILED_FILE_LIST_NAME)
                FileUtils.save_data_to_file(error_file_path, "%s\n" % result[self.MULTI_THREAD_ERROR_FILE_INDEX],
                                            'a+', delete=False)

    def _do_multi_process(self: any, file_list: list) -> int:
        cpu_count = int((multiprocessing.cpu_count() + 1) / 2)
        pool = multiprocessing.Pool(cpu_count)
        all_task = []
        for cur_path in file_list:
            task = pool.apply_async(self._process_func,
                                    args=(cur_path,),
                                    callback=self._handle_result_callback)
            all_task.append(task)
        pool.close()
        pool.join()
        for task in all_task:
            result = task.get()
            if len(result) != self.MULTI_THREAD_RESULT_COUNT:
                continue
            cur_ret = result[self.MULTI_THREAD_RETURN_CODE_INDEX]
            if cur_ret != CompareError.MSACCUCMP_NONE_ERROR:
                return cur_ret
        return CompareError.MSACCUCMP_NONE_ERROR

    def _split_big_file(self: any) -> (list, list):
        if len(self._input_path) == 1:
            files = os.listdir(self._input_path[0])
        else:
            files = self._input_path
        self._progress = Progress(len(files))
        multi_process_file_list = []
        big_file_list = []
        max_file_size = self.get_max_file_size()
        for cur_file in files:
            if cur_file == ConstManager.MAPPING_FILE_NAME:
                continue
            if len(self._input_path) == 1:
                cur_path = os.path.join(self._input_path[0], cur_file)
            else:
                cur_path = cur_file
            if os.path.isfile(cur_path):
                # skip big file
                if os.path.getsize(cur_path) > max_file_size:
                    big_file_list.append(cur_path)
                else:
                    multi_process_file_list.append(cur_path)
        return multi_process_file_list, big_file_list



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from tqdm import tqdm

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.utils import logger, CompareException
from msprobe.core.common.file_utils import load_yaml
from msprobe.core.compare.config import ModeConfig
from msprobe.core.compare.utils import gen_api_batches


cur_dir = os.path.dirname(os.path.realpath(__file__))
diff_threshold_yaml_path = os.path.join(cur_dir, 'diff_analyze_threshold.yaml')
ignore_op_list_yaml_path = os.path.join(cur_dir, 'ignore_op_list.yaml')
ignore_list = load_yaml(ignore_op_list_yaml_path)
thresholds = load_yaml(diff_threshold_yaml_path)
cmp_metrics = thresholds.get('compare_metrics')


class FirstDiffAnalyze:
    def __init__(self, mode_config: ModeConfig, rank):
        self.mode_config = mode_config
        self.rank = rank

    @staticmethod
    def single_metric_diff_check(cmp_metric, metric_value):
        threshold = thresholds.get(cmp_metric, None)
        if threshold is None:
            logger.error(f"Check diff or {cmp_metric} need to configure the threshold. "
                         f"Please configure it in 'diff_analyze_threshold.yaml'.")
            raise CompareException(CompareException.MISSING_THRESHOLD_ERROR)
        if not isinstance(threshold, list) or len(threshold) != 1:
            logger.error(f"{cmp_metric} threshold configure wrong. Please check.")
            raise CompareException(CompareException.WRONG_THRESHOLD_ERROR)
        if isinstance(metric_value, str) and metric_value.endswith('%'):
            metric_value_float = float(metric_value[:-1]) / 100
            if metric_value_float > threshold[0]:
                return True
        return False

    def single_api_check(self, result_slice, header, api_name=None):
        """
        单个api差异检查

        :param result_slice: 数据切片
        :param header: 列名列表
        :return: {'is_same': bool, 'op_items': list[dict]}
        """
        single_check_result = {
            'is_same': True,
            'op_items': []
        }

        column_indices = {name: idx for idx, name in enumerate(header)}
        output_idx = -1
        for line in result_slice:
            op_item = {
                column_name: line[column_indices[column_name]]
                for column_name in header
            }
            single_check_result['op_items'].append(op_item)
            if op_item['state'] != 'output':
                continue
            output_idx += 1
            if output_idx in ignore_list.get(api_name, []):
                continue
            # set is_same
            if self.mode_config.dump_mode == Const.MD5:
                if line[column_indices[CompareConst.RESULT]] == CompareConst.DIFF:
                    single_check_result['is_same'] = False
            else:
                for cmp_metric in cmp_metrics:
                    metric_value = line[column_indices[cmp_metric]]
                    if self.single_metric_diff_check(cmp_metric, metric_value):
                        single_check_result['is_same'] = False
                        break
        return single_check_result

    def check(self, result_df):
        """
        比对后循环遍历api检查差异
        example：
        {
            'Functional.conv2d.0.forward': {
                'is_same': true,
                'op_items': [
                    {
                        'NPU name': 'Functional.conv2d.0.forward.input.0',
                        'Bench name': 'Functional.conv2d.0.forward.input.0',
                        'xxx': 1,
                        'NormRelativeErr': 2,
                        'yyy': 3,
                        ...
                    }
                ]
            }
        }
        """
        result = result_df.values
        header = result_df.columns.tolist()

        api_batches = gen_api_batches(result, header)

        check_result = {}

        default_bar_desc = 'API/Module diff check Progress'
        bar_desc_add_rank = f'[{self.rank}]' + default_bar_desc if self.rank else default_bar_desc
        with tqdm(total=len(api_batches), desc=bar_desc_add_rank, unit="api/module", ncols=100) as progress_bar:
            for api_batch in api_batches:
                result_slice = result[api_batch.start: api_batch.params_grad_end_index]
                api_compo = api_batch.api_name.split('.')
                # suppose name is Tensor.MatMul.0.forward
                if len(api_compo) < 4:
                    continue
                # get MatMul as api_name
                api_name = api_compo[-3]
                check_result[api_batch.api_name] = self.single_api_check(result_slice, header, api_name)
                progress_bar.update(1)

        return check_result



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
RangeMode class.
This class mainly involves functions for selecting operators.
"""

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_rule import CompareRule
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.vector_cmp.range_manager import range_manager


class RangeMode(range_manager.RangeManager):
    """
    The subclass of range manager
    """

    def __init__(self: any, input_str: str) -> None:
        super(range_manager.RangeManager, self).__init__()
        self.start, self.end, self.step = self._parse_input_str(input_str)

    @staticmethod
    def _parse_input_str(input_str: str) -> (int, int, int):
        cur_range = [ConstManager.DEFAULT_START, ConstManager.DEFAULT_END, ConstManager.DEFAULT_STEP]
        range_list = input_str.split(',')
        if len(range_list) != len(cur_range):
            log.print_error_log('The range (%s) is invalid, just supports "start,end,step".' % input_str)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        for index, item in enumerate(range_list):
            value = item.strip()
            if not value:
                continue
            if index == ConstManager.END_INDEX and value == '-1':
                continue
            if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, value):
                log.print_error_log('The range (%s) is invalid, just supports '
                                    '"start,end,step", the value is number.' % input_str)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            cur_range[index] = int(value)
        return (x for x in cur_range)

    def get_all_ops(self: any, compare_rule: CompareRule) -> list:
        """
        Get all operators according to [start, end, step]:
        :param compare_rule: the compare rule
        :return: operator list
        """
        op_list = []
        for op in compare_rule.fusion_info.op_list:
            op_sequence = op.attr.get_op_sequence()
            if op_sequence > self.end:
                break
            if op_sequence < self.start:
                continue
            if self.start == op_sequence or self.end == op_sequence or ((op_sequence - self.start) % self.step) == 0:
                op_list.append(op)
        return self._get_op_list(op_list, compare_rule)

    def check_input_valid(self: any, op_count: int) -> None:
        """
        Check range valid:
        :param op_count: the op count
        """
        if self.start < 1 or self.start > op_count:
            log.print_out_of_range_error('', 'start', self.start, '[1, %d]' % op_count)
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
        if self.end == -1:
            self.end = op_count
        if self.end < self.start or self.end > op_count:
            log.print_out_of_range_error('', 'end', self.end, '[1, %d]' % op_count)
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
        if self.step < 1 or self.step >= op_count:
            log.print_out_of_range_error('', 'step', self.step, '[1, %d)' % op_count)
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
        log.print_info_log('The range compare for [%d,%d,%d].' % (self.start, self.end, self.step))



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.file_utils import create_directory, check_path_readability, check_path_writability
from msprobe.core.common.const import Const
from msprobe.core.common.log import logger



def try_import_torchair():
    try:
        import torch
        import torch_npu
        import torchair
    except ModuleNotFoundError as e:
        logger.error("torch or torch_npu with torchair not found. Try installing the latest torch_npu.")
        raise e


def set_ge_dump_config(
        dump_path='',
        dump_mode='all',
        fusion_switch_file=None,
        dump_token=None,
        dump_layer=None,
        compiler_config=None
):
    try_import_torchair()
    from torchair.configs.compiler_config import CompilerConfig

    dump_path = os.path.realpath(os.path.join(dump_path, Const.TORCHAIR_GE_DATA_DIRECTORY))
    create_directory(dump_path)
    check_path_readability(dump_path)
    check_path_writability(dump_path)
    if os.listdir(dump_path):
        logger.warning('The specified directory for GE-Dump data is not empty, which may result in data mixing.')

    if compiler_config is not None and not isinstance(compiler_config, CompilerConfig):
        raise TypeError(f'compiler_config must be a CompilerConfig, but got {type(compiler_config)}')
    config = compiler_config if compiler_config else CompilerConfig()
    # Generate GE mapping graph
    config.debug.graph_dump.type = "txt"
    if hasattr(config.debug.graph_dump, "_path"):  # interface changed since 8.0.RC1.b080
        setattr(config.debug.graph_dump, "_path", dump_path)
    else:
        config.debug.graph_dump.path = dump_path

    if fusion_switch_file is not None:
        if not os.path.exists(fusion_switch_file):
            raise FileNotFoundError(f'fusion_switch_file: {fusion_switch_file} not found')
        config.fusion_config.fusion_switch_file = fusion_switch_file

    # Enable GE dump
    config.dump_config.enable_dump = True
    config.dump_config.dump_mode = dump_mode
    config.dump_config.dump_path = dump_path

    if dump_token is not None:
        new_token = [str(x) for x in dump_token]
        config.dump_config.dump_step = "|".join(new_token)
    if dump_layer is not None:
        dump_layer = " ".join(dump_layer)
        config.dump_config.dump_layer = dump_layer

    return config


def set_fx_dump_config(dump_path='', op_list=None, compiler_config=None):
    try_import_torchair()
    from torchair.configs.compiler_config import CompilerConfig

    dump_path = os.path.realpath(os.path.join(dump_path, Const.TORCHAIR_FX_DATA_DIRECTORY))
    create_directory(dump_path)
    check_path_readability(dump_path)
    check_path_writability(dump_path)
    if os.listdir(dump_path):
        logger.warning('The specified directory for FX-Dump data is not empty, which may result in data mixing.')

    if compiler_config is not None and not isinstance(compiler_config, CompilerConfig):
        raise TypeError(f'compiler_config must be a CompilerConfig, but got {type(compiler_config)}')
    config = compiler_config if compiler_config else CompilerConfig()
    if op_list:
        if not isinstance(op_list, list) or not all(isinstance(item, str) for item in op_list):
            raise TypeError("op_list must be a list of strings")
        from msprobe.infer.utils.util import filter_cmd
        op_list = filter_cmd(op_list)
        config.debug.data_dump.filter = (
            lambda x: x
            if any(keyword.lower() in x.name.lower() for keyword in op_list)
            else None
        )
        

    # Enable FX dump
    config.debug.data_dump.type = "npy"
    if hasattr(config.debug.data_dump, "path"):
        config.debug.data_dump.path = dump_path

    return config



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
RemoveInplaceLayerProcess class.
This class mainly involves the remove_inplace_layer function.
"""

import sys
import stat
import argparse
import os

import google.protobuf.text_format
import caffe.proto.caffe_pb2 as caffe_pb2

from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils import log, file_utils
from msprobe.msaccucmp.cmp_utils import path_check
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string

MAX_SIZE = 10 * 1024 * 1024 * 1024


class RemoveInplaceLayerProcess:
    """
    The class for remove inplace layer for caffe prototxt
    """

    WRITE_FLAGS = os.O_WRONLY | os.O_CREAT
    WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR

    def __init__(self: any) -> None:
        parse = argparse.ArgumentParser()
        parse.add_argument("-i", dest="input_file_path", help="<Required> the prototxt file path",
                           type=safe_path_string, required=True)
        parse.add_argument("-o", dest="output_file_path", help="<Optional> the output file path", type=safe_path_string)
        args, _ = parse.parse_known_args(sys.argv[1:])
        self.input_file_path = os.path.realpath(args.input_file_path)
        if args.output_file_path:
            if os.path.islink(os.path.abspath(args.output_file_path)):
                log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_file_path)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
            self.output_file_path = os.path.realpath(args.output_file_path)
        else:
            output_file_path = os.path.join(
                os.path.dirname(self.input_file_path),
                "new_" + os.path.basename(self.input_file_path))
            self.output_file_path = os.path.realpath(output_file_path)
        self.net_param = None
        self.cur_layer_idx = -1

    @staticmethod
    def _check_input_file_valid(path: str) -> None:
        ret = path_check.check_path_valid(
            path, exist=True, have_write_permission=False, path_type=path_check.PathType.File
        )
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

        if os.path.getsize(path) > MAX_SIZE:
            log.print_error_log("The file '%r' is too large." % path)
            raise CompareError(CompareError.MSACCUCMP_FILE_TOO_LARGE_ERROR)

    @staticmethod
    def _check_output_file_valid(path: str) -> None:
        ret = path_check.check_path_valid(
            path, exist=False, have_write_permission=True, path_type=path_check.PathType.File
        )
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

        if os.path.exists(path):
            if not os.path.isfile(path):
                log.print_error_log("Provided output_file_path exists but is not a file.")
                raise CompareError(CompareError.MSACCUCMP_SYMLINK_ERROR)
            os.remove(path)
            log.print_warn_log("The file '%r' already exists" % path)

    def check_arguments_valid(self: any) -> None:
        """
        Check file valid.
        """
        self._check_input_file_valid(self.input_file_path)
        self._check_output_file_valid(self.output_file_path)

    def remove_inplace_layer(self: any) -> None:
        """
        remove inplace layer and save new layer to file
        """
        # check path valid
        self.check_arguments_valid()

        # read prototxt file
        self.net_param = caffe_pb2.NetParameter()
        try:
            with open(self.input_file_path, 'rb') as model_file:
                google.protobuf.text_format.Parse(model_file.read(), self.net_param)
        except (google.protobuf.text_format.ParseError, UnicodeDecodeError) as error:
            log.print_error_log("Provided input_file_path is not a valid protobuf text file")
            raise CompareError(CompareError.MSACCUCMP_INVALID_FORMAT_ERROR) from error

        # parse net
        while True:
            old_name, new_name = self._find_name()
            if not old_name and not new_name:
                break
            self._parse_name(old_name, new_name)

        # remove Dropout type
        for (_, layer_item) in enumerate(self.net_param.layer):
            if layer_item.type == 'Dropout':
                self.net_param.layer.remove(layer_item)

        # write file to new path
        with os.fdopen(os.open(self.output_file_path, self.WRITE_FLAGS, self.WRITE_MODES), 'w') as open_file:
            file_content = str(self.net_param)
            open_file.write(file_content)
        log.print_info_log('The "%r" has removed inplace layer.' % self.input_file_path)
        log.print_info_log('The new prototxt file has been saved to "%r".' % self.output_file_path)

    def _handle_top(self: any, layer_item: any, layer_idx: int) -> (bool, str, str):
        for (top_index, top_item) in enumerate(layer_item.top):
            if layer_item.type == 'Dropout':
                if len(layer_item.bottom) != 1:
                    return True, '', ''
                bottom_item = layer_item.bottom[0]
                if top_item != bottom_item:
                    layer_item.top[top_index] = bottom_item
                    old_name = top_item
                    new_name = bottom_item
                    self.cur_layer_idx = layer_idx
                    return True, old_name, new_name
            elif top_item != layer_item.name:
                if len(layer_item.top) == 1 and len(layer_item.bottom) == 1 \
                        and layer_item.top[0] == layer_item.bottom[0]:
                    layer_item.top[top_index] = layer_item.name
                    old_name = top_item
                    new_name = layer_item.name
                    self.cur_layer_idx = layer_idx
                    return True, old_name, new_name
        return False, '', ''

    def _find_name(self: any) -> (str, str):
        for (layer_idx, layer_item) in enumerate(self.net_param.layer):
            if layer_idx <= self.cur_layer_idx:
                continue
            ok, old_name, new_name = self._handle_top(layer_item, layer_idx)
            if ok:
                return old_name, new_name
        return '', ''

    def _parse_name(self: any, old_name: str, new_name: str) -> None:
        for (layer_idx, layer_item) in enumerate(self.net_param.layer):
            if layer_idx <= self.cur_layer_idx:
                continue
            for (bottom_index, bottom_item) in enumerate(layer_item.bottom):
                if bottom_item == old_name:
                    layer_item.bottom[bottom_index] = new_name
            for (top_index, top_item) in enumerate(layer_item.top):
                if top_item == old_name:
                    layer_item.top[top_index] = new_name


if __name__ == "__main__":
    with file_utils.UmaskWrapper():
        try:
            RemoveInplaceLayerProcess().remove_inplace_layer()
        except Exception as base_err:
            log.print_error_log(f'Basic error running {sys.argv[0]}: {base_err}')
            sys.exit(1)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.core.dump.common_config import BaseConfig, CommonConfig


class TensorConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.check_mode = None
        self.file_format = json_config.get("file_format")
        self.check_config()
        self._check_summary_mode()
        self._check_config()

    def _check_config(self):
        if self.file_format and self.file_format not in ["npy", "bin"]:
            raise Exception("file_format is invalid")


class StatisticsConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.file_format = None
        self.check_mode = None
        self.check_config()
        self._check_summary_mode()

        self.tensor_list = json_config.get("tensor_list", [])
        self._check_str_list_config(self.tensor_list, "tensor_list")
        self.stat_cal_mode = json_config.get("device", "host")
        self.device_stat_precision_mode = json_config.get("precision", "high")
        self._check_stat_params()

    def _check_stat_params(self):
        if self.stat_cal_mode not in ["device", "host"]:
            raise Exception("Config param [device] is invalid, expected from [\"device\", \"host\"]")
        if self.device_stat_precision_mode not in ["high", "low"]:
            raise Exception("Config param [precision] is invalid, expected from [\"high\", \"low\"]")

    def _check_summary_mode(self):
        muti_opt = ["max", "min", "mean", "count", "negative zero count", "positive zero count", "nan count",
                    "negative inf count", "positive inf count", "zero count", "l2norm", "hash", "md5"]
        if isinstance(self.summary_mode, str) and self.summary_mode not in Const.SUMMARY_MODE:
            raise Exception("summary_mode is an invalid string")
        if isinstance(self.summary_mode, list) and not all(opt in muti_opt for opt in self.summary_mode):
            raise Exception("summary_mode contains invalid option(s)")


class OverflowCheckConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.data_mode = ["all"]
        self._check_config()

    def _check_config(self):
        if self.check_mode and self.check_mode not in ["all", "aicore", "atomic"]:
            raise Exception("check_mode is invalid")


class ExceptionDumpConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.data_mode = ["all"]


class StructureConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)


TaskDict = {
    Const.TENSOR: TensorConfig,
    Const.STATISTICS: StatisticsConfig,
    Const.OVERFLOW_CHECK: OverflowCheckConfig,
    Const.STRUCTURE: StructureConfig,
    Const.EXCEPTION_DUMP: ExceptionDumpConfig
}


def parse_common_config(json_config):
    return CommonConfig(json_config)


def parse_task_config(task, json_config):
    task_map = json_config.get(task)
    if not task_map:
        task_map = dict()
    if task not in TaskDict:
        raise Exception("task is invalid.")
    return TaskDict.get(task)(task_map)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import List

from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException


class RankGroupGenerator(object):
    def __init__(self, tensor_parallel: int, expert_parallel: int, data_parallel: int,
                 pipeline_parallel: int, context_parallel: int, order: str) -> None:
        self.tensor_parallel = tensor_parallel
        self.expert_parallel = expert_parallel
        self.data_parallel = data_parallel
        self.pipeline_parallel = pipeline_parallel
        self.context_parallel = context_parallel
        self.total_size = tensor_parallel * data_parallel * pipeline_parallel * context_parallel

        self.parallel_sizes = {
            "tp": self.tensor_parallel,
            "pp": self.pipeline_parallel,
            "dp": self.data_parallel,
            "ep": self.expert_parallel,
            "cp": self.context_parallel,
        }
        self.original_order = order
        normalized_order = order.lower()

        # 检查ep和dp是否相邻
        if 'ep' in normalized_order:
            if 'ep-dp' not in normalized_order and 'dp-ep' not in normalized_order:
                logger.error(f"The ep and dp must be adjacent in order ({self.original_order}).")
                raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

        # 检查所有非1的并行维度是否都在order中
        for name in self.parallel_sizes.keys():
            size = self.parallel_sizes[name]
            if name not in normalized_order:
                if size != 1:
                    logger.error(f"The parallel size ({name}) is ({size}), "
                                 f"but it's not specified in order ({self.original_order}).")
                    raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
                else:
                    normalized_order += '-' + name

        self.order_with_ep = normalized_order
        self.order_without_ep = '-'.join([item for item in normalized_order.split('-') if item != 'ep'])

        self.size_list_with_ep = []
        self.size_list_without_ep = []

        for item in normalized_order.split('-'):
            if item == 'dp':
                self.size_list_with_ep.append(self.data_parallel // self.expert_parallel)
                self.size_list_without_ep.append(self.data_parallel)
            elif item == 'ep':
                self.size_list_with_ep.append(self.expert_parallel)
            else:
                self.size_list_with_ep.append(self.parallel_sizes[item])
                self.size_list_without_ep.append(self.parallel_sizes[item])

    @staticmethod
    def create_mask(order_str: str, target_tokens: str) -> List[bool]:
        order_elements = order_str.split('-')
        target_elements = target_tokens.split('-')
        mask = [False] * len(order_elements)
        for token in target_elements:
            mask[order_elements.index(token)] = True
        return mask

    @staticmethod
    def create_masked_rank_groups(
            total_size: int,
            parallel_dims: List[int],
            mask: List[bool],
    ) -> List[List[int]]:
        def compute_prefix_products(dimensions: List[int], initial: int = 1) -> List[int]:
            products = [initial]
            current = initial
            for dim in dimensions:
                current *= dim
                products.append(current)
            return products

        def calculate_inner_product(a: List[int], b: List[int]) -> int:
            return sum(x * y for x, y in zip(a, b))

        def decompose_index(index: int, shape: List[int], strides: List[int] = None) -> List[int]:
            if strides is None:
                strides = compute_prefix_products(shape)
            indices = [(index // stride) % dim for dim, stride in zip(shape, strides)]

            # 验证分解是否正确
            if calculate_inner_product(indices, strides[:-1]) != index:
                error_msg = f"The index {index} with shape {shape} doesn't match decomposed indices {indices}."
                logger.error(error_msg)
                raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

            return indices

        # 分离被掩码和未被掩码的维度
        masked_dims = [dim for dim, is_masked in zip(parallel_dims, mask) if is_masked]
        unmasked_dims = [dim for dim, is_masked in zip(parallel_dims, mask) if not is_masked]

        # 计算全局、掩码和未掩码的步长
        global_strides = compute_prefix_products(parallel_dims)
        masked_strides = [stride for stride, is_masked in zip(global_strides, mask) if is_masked]
        unmasked_strides = [stride for stride, is_masked in zip(global_strides, mask) if not is_masked]

        # 计算组大小和组数
        group_dim = compute_prefix_products(masked_dims)[-1]
        group_count = total_size // group_dim

        # 生成所有组的rank
        rank_groups = []
        for group_idx in range(group_count):
            decomposed_group = decompose_index(group_idx, unmasked_dims)
            current_group = []
            for in_group_idx in range(group_dim):
                decomposed_rank = decompose_index(in_group_idx, masked_dims)
                rank_value = (calculate_inner_product(decomposed_rank, masked_strides) +
                              calculate_inner_product(decomposed_group, unmasked_strides))
                current_group.append(rank_value)
            rank_groups.append(current_group)

        return rank_groups

    def generate_ranks(self, token: str, separate_ep: bool = False) -> List[List[int]]:
        if separate_ep:
            parallel_dims = self.size_list_with_ep
            current_order = self.order_with_ep
        else:
            parallel_dims = self.size_list_without_ep
            current_order = self.order_without_ep

        mask = self.create_mask(current_order, token)
        return self.create_masked_rank_groups(self.total_size, parallel_dims, mask)

    def generate_all_ranks(self) -> dict:
        result = {}
        for token in ["dp", "pp", "tp"]:
            result[token] = self.generate_ranks(token)
            result[f"{token}_size"] = self.parallel_sizes[token]
        return result


def get_tp_pp_default_groups(
        total_world_size: int,
        tensor_parallel_size: int = 1,
        pipeline_parallel_size: int = 1,
        order: str = "tp-cp-ep-dp-pp",
) -> tuple:
    context_parallel_size = 1
    expert_parallel_size = 1

    # 检查world_size是否可被各并行维度的乘积整除
    product = tensor_parallel_size * pipeline_parallel_size * context_parallel_size
    if total_world_size % product != 0:
        logger.error(f"The world size ({total_world_size}) is not divisible by "
                     f"{tensor_parallel_size} x {pipeline_parallel_size} x {context_parallel_size}.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    data_parallel_size = total_world_size // product

    # 检查数据并行是否可被专家并行整除
    if data_parallel_size % expert_parallel_size != 0:
        logger.error(f"The data parallel size ({data_parallel_size}) is not divisible by expert parallel size.")
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    # 生成rank组
    rank_creator = RankGroupGenerator(
        tensor_parallel=tensor_parallel_size,
        expert_parallel=expert_parallel_size,
        data_parallel=data_parallel_size,
        pipeline_parallel=pipeline_parallel_size,
        context_parallel=context_parallel_size,
        order=order,
    )

    return rank_creator.generate_ranks('tp'), rank_creator.generate_ranks('pp')



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the const value.
"""
import os
import stat
import numpy as np


class DD:
    # Format 类型
    FORMAT_NCHW = 0
    FORMAT_NHWC = 1
    FORMAT_ND = 2
    FORMAT_NC1HWC0 = 3
    FORMAT_FRACTAL_Z = 4
    FORMAT_NC1C0HWPAD = 5
    FORMAT_NHWC1C0 = 6
    FORMAT_FSR_NCHW = 7
    FORMAT_FRACTAL_DECONV = 8
    FORMAT_C1HWNC0 = 9
    FORMAT_FRACTAL_DECONV_TRANSPOSE = 10
    FORMAT_FRACTAL_DECONV_SP_STRIDE_TRANS = 11
    FORMAT_NC1HWC0_C04 = 12
    FORMAT_FRACTAL_Z_C04 = 13
    FORMAT_CHWN = 14
    FORMAT_FRACTAL_DECONV_SP_STRIDE8_TRANS = 15
    FORMAT_HWCN = 16
    FORMAT_NC1KHKWHWC0 = 17
    FORMAT_BN_WEIGHT = 18
    FORMAT_FILTER_HWCK = 19
    FORMAT_HASHTABLE_LOOKUP_LOOKUPS = 20
    FORMAT_HASHTABLE_LOOKUP_KEYS = 21
    FORMAT_HASHTABLE_LOOKUP_VALUE = 22
    FORMAT_HASHTABLE_LOOKUP_OUTPUT = 23
    FORMAT_HASHTABLE_LOOKUP_HITS = 24
    FORMAT_C1HWNCoC0 = 25
    FORMAT_MD = 26
    FORMAT_NDHWC = 27
    FORMAT_FRACTAL_ZZ = 28
    FORMAT_FRACTAL_NZ = 29
    FORMAT_NCDHW = 30
    FORMAT_DHWCN = 31
    FORMAT_NDC1HWC0 = 32
    FORMAT_FRACTAL_Z_3D = 33
    FORMAT_CN = 34
    FORMAT_NC = 35
    FORMAT_DHWNC = 36
    FORMAT_FRACTAL_Z_3D_TRANSPOSE = 37
    FORMAT_FRACTAL_ZN_LSTM = 38
    FORMAT_FRACTAL_Z_G = 39
    FORMAT_RESERVED = 40
    FORMAT_ALL = 41
    FORMAT_NULL = 42
    FORMAT_ND_RNN_BIAS = 43
    FORMAT_FRACTAL_ZN_RNN = 44
    FORMAT_YUV = 45
    FORMAT_YUV_A = 46
    FORMAT_MAX = 0xff

    # DataType 类型
    DT_UNDEFINED = 0
    DT_FLOAT = 1
    DT_FLOAT16 = 2
    DT_INT8 = 3
    DT_UINT8 = 4
    DT_INT16 = 5
    DT_UINT16 = 6
    DT_INT32 = 7
    DT_INT64 = 8
    DT_UINT32 = 9
    DT_UINT64 = 10
    DT_BOOL = 11
    DT_DOUBLE = 12
    DT_STRING = 13
    DT_DUAL_SUB_INT8 = 14
    DT_DUAL_SUB_UINT8 = 15
    DT_COMPLEX64 = 16
    DT_COMPLEX128 = 17
    DT_QINT8 = 18
    DT_QINT16 = 19
    DT_QINT32 = 20
    DT_QUINT8 = 21
    DT_QUINT16 = 22
    DT_RESOURCE = 23
    DT_STRING_REF = 24
    DT_DUAL = 25
    DT_BF16 = 27
    DT_UINT1 = 29
    DT_COMPLEX32 = 32
    DT_HIFLOAT8 = 33
    DT_FLOAT8_E5M2 = 34
    DT_FLOAT8_E4M3FN = 35
    DT_FLOAT8_E8M0 = 36
    DT_FLOAT6_E3M2 = 37
    DT_FLOAT6_E2M3 = 38
    DT_FLOAT4_E2M1 = 39
    DT_FLOAT4_E1M2 = 40
    # 其他
    L1 = 0


class ConstManager:
    """
    The class for const manager
    """
    # common const
    LINUX_FILE_NAME_MAX_LEN = 255
    LINUX_PATH_MAX_LEN = 4096
    INPUT = 'input'
    OUTPUT = 'output'
    SUPPORT_DETAIL_TYPE = [INPUT, OUTPUT]
    FOUR_DIMS_LENGTH = 4
    FIVE_DIMS_LENGTH = 5
    WRITE_FLAGS = os.O_WRONLY | os.O_CREAT
    WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR
    OLD_DUMP_TYPE = 0
    PROTOBUF_DUMP_TYPE = 1
    BINARY_DUMP_TYPE = 2
    N0_AXIS = 16
    C0_AXIS = 16

    TIMESTAMP_LENGTH = 16
    INVALID_TIMESTAMP = 0
    INVALID_FILE_TYPE = 0
    TRAD_MODE = 0


    # overflow
    OVERFLOW_MIN_VALUE = 5.96e-8
    OVERFLOW_MAX_VALUE = 65504
    MAGIC_KEY_WORD = 'magic'
    MAGIC_NUM = 0x5a5a5a5a
    INVALID_ID = 65535
    STATUS_LEN = 4

    # algorithm
    FLOAT_EPSILON = np.finfo(float).eps
    NAN = 'NaN'

    # progress const
    MAX_PROGRESS_LINE_COUNT = 100000
    MAX_PROGRESS = 100

    TIME_LENGTH = 1000
    DELIMITER = ':'

    # special ops
    SPECIAL_OPS_TYPE = ("Cast", "TransData")

    # single op const

    DEFAULT_TOP_N = 20
    MAX_DETAIL_INFO_LINE_COUNT = 1000000
    DETAIL_LINE_COUNT_RANGE_MAX = 1000000
    DETAIL_LINE_COUNT_RANGE_MIN = 10000
    CSV_SUFFIX = ".csv"
    TXT_SUFFIX = ".txt"
    NPY_SUFFIX = '.npy'
    SUMMARY_TXT_SUFFIX = "_summary.txt"
    SIMPLE_OP_MAPPING_FILE_NAME = "simple_op_mapping.csv"

    FUSION_OP_INDEX = 1

    # algorithm manager const
    BUILT_IN_ALGORITHM_DIR_NAME = "builtin_algorithm"
    CUSTOM_ALGORITHM_DIR_NAME = "algorithm"
    COMPARE_FUNC_NAME = 'compare'
    BUILT_IN_ALGORITHM = [
        "CosineSimilarity",
        "MaxAbsoluteError",
        "AccumulatedRelativeError",
        "RelativeEuclideanDistance",
        "KullbackLeiblerDivergence",
        "StandardDeviation",
        "MeanAbsoluteError",
        "RootMeanSquareError",
        "MaxRelativeError",
        "MeanRelativeError"
    ]
    ALGORITHM_FILE_NAME_PATTERN = r"^alg_([A-Za-z0-9_]+)\.py[c]?$"
    BOOL_ALGORITHM = ('CosineSimilarity', 'RelativeEuclideanDistance')
    FILE_CMP_SUPPORTED_ALGORITHM = [
        "CosineSimilarity",
        "MaxAbsoluteError",
        "AccumulatedRelativeError",
        "RelativeEuclideanDistance",
        "MeanAbsoluteError",
        "RootMeanSquareError",
        "MaxRelativeError",
        "MeanRelativeError"
    ]

    COMPARE_ARGUMENT_COUNT = 3
    SPACE = ' '
    BUILTIN = 'Builtin'
    CUSTOM = 'Custom'

    # built in algorithm const
    MINIMUM_VALUE = 0.001

    INPUT_OP_NAME_INDEX = 0
    INPUT_INDEX_INDEX = 1
    INPUT_INFO_COUNT = 2

    TENSOR_INDEX = 7
    INPUT_PATTERN = ":input:"
    OUTPUT_PATTERN = ":output:"

    # fusion rule const
    GRAPH_OBJECT = "graph"
    OP_OBJECT = "op"
    NAME_OBJECT = "name"
    TYPE_OBJECT = "type"
    ID_OBJECT = 'id'
    INPUT_OBJECT = "input"
    ATTR_OBJECT = "attr"
    L1_FUSION_SUB_GRAPH_NO_OBJECT = "_L1_fusion_sub_graph_no"
    ORIGINAL_OP_NAMES_OBJECT = "_datadump_original_op_names"
    OUTPUT_DESC_OBJECT = "output_desc"
    ORIGIN_NAME_OBJECT = "_datadump_origin_name"
    ORIGIN_OUTPUT_INDEX_OBJECT = "_datadump_origin_output_index"
    ORIGIN_FORMAT_OBJECT = "_datadump_origin_format"
    IS_MULTI_OP = "_datadump_is_multiop"
    GE_ORIGIN_FORMAT_OBJECT = "origin_format"
    GE_ORIGIN_SHAPE_OBJECT = "origin_shape"
    D_TYPE = "dtype"
    KEY_OBJECT = "key"
    VALUE_OBJECT = "value"
    STRING_TYPE_OBJECT = "s"
    INT_TYPE_OBJECT = "i"
    BOOL_TYPE_OBJECT = 'b'
    LIST_TYPE_OBJECT = "list"
    DATA_OBJECT = "Data"

    # network compare range
    START_INDEX = 1
    END_INDEX = 1
    STEP_INDEX = 1
    DEFAULT_START = 1
    DEFAULT_END = -1
    DEFAULT_STEP = 1
    OP_SEQUENCE_INDEX = 1
    OP_SEQUENCE = 'OpSequence'
    RANGE_MANAGER_KEY = 'range_manager'
    RANGE_MODE = 'range'
    SELECT_MODE = 'select'

    # dump file
    LEFT_TYPE = 'Left'
    RIGHT_TYPE = 'Right'

    VECTOR_COMPARE_HEADER = [
        "OpType", "NPUDump", "DataType", "Address",
        "GroundTruth", "DataType", "Address", "TensorIndex"
    ]
    MY_OUTPUT_ADDRESS_INDEX = 3
    GROUND_TRUTH_ADDRESS_INDEX = 6

    MAPPING_FILE_HEADER = [
        "Index", "OpType", "NPUDump", "GroundTruth", "TensorIndex", "NPUDumpPath", "GroundTruthPath"
    ]

    STRING_TO_FORMAT_MAP = {
        "NCHW": DD.FORMAT_NCHW,
        "NHWC": DD.FORMAT_NHWC,
        "ND": DD.FORMAT_ND,
        "NC1HWC0": DD.FORMAT_NC1HWC0,
        "FRACTAL_Z": DD.FORMAT_FRACTAL_Z,
        "NC1C0HWPAD": DD.FORMAT_NC1C0HWPAD,
        "NHWC1C0": DD.FORMAT_NHWC1C0,
        "FSR_NCHW": DD.FORMAT_FSR_NCHW,
        "FRACTAL_DECONV": DD.FORMAT_FRACTAL_DECONV,
        "C1HWNC0": DD.FORMAT_C1HWNC0,
        "FRACTAL_DECONV_TRANSPOSE": DD.FORMAT_FRACTAL_DECONV_TRANSPOSE,
        "FRACTAL_DECONV_SP_STRIDE_TRANS": DD.FORMAT_FRACTAL_DECONV_SP_STRIDE_TRANS,
        "NC1HWC0_C04": DD.FORMAT_NC1HWC0_C04,
        "FRACTAL_Z_C04": DD.FORMAT_FRACTAL_Z_C04,
        "CHWN": DD.FORMAT_CHWN,
        "DECONV_SP_STRIDE8_TRANS": DD.FORMAT_FRACTAL_DECONV_SP_STRIDE8_TRANS,
        "NC1KHKWHWC0": DD.FORMAT_NC1KHKWHWC0,
        "BN_WEIGHT": DD.FORMAT_BN_WEIGHT,
        "FILTER_HWCK": DD.FORMAT_FILTER_HWCK,
        "HWCN": DD.FORMAT_HWCN,
        "LOOKUP_LOOKUPS": DD.FORMAT_HASHTABLE_LOOKUP_LOOKUPS,
        "LOOKUP_KEYS": DD.FORMAT_HASHTABLE_LOOKUP_KEYS,
        "LOOKUP_VALUE": DD.FORMAT_HASHTABLE_LOOKUP_VALUE,
        "LOOKUP_OUTPUT": DD.FORMAT_HASHTABLE_LOOKUP_OUTPUT,
        "LOOKUP_HITS": DD.FORMAT_HASHTABLE_LOOKUP_HITS,
        "MD": DD.FORMAT_MD,
        "NDHWC": DD.FORMAT_NDHWC,
        "C1HWNCoC0": DD.FORMAT_C1HWNCoC0,
        "FRACTAL_NZ": DD.FORMAT_FRACTAL_NZ,
        "NCDHW": DD.FORMAT_NCDHW,
        "DHWCN": DD.FORMAT_DHWCN,
        "NDC1HWC0": DD.FORMAT_NDC1HWC0,
        "FRACTAL_Z_3D": DD.FORMAT_FRACTAL_Z_3D,
        "CN": DD.FORMAT_CN,
        "DHWNC": DD.FORMAT_DHWNC,
        "FRACTAL_Z_3D_TRANSPOSE": DD.FORMAT_FRACTAL_Z_3D_TRANSPOSE,
        "FRACTAL_ZN_LSTM": DD.FORMAT_FRACTAL_ZN_LSTM,
        "FRACTAL_Z_G": DD.FORMAT_FRACTAL_Z_G,
        "RESERVED": DD.FORMAT_RESERVED,
        "ALL": DD.FORMAT_ALL,
        "NULL": DD.FORMAT_NULL,
        "ND_RNN_BIAS": DD.FORMAT_ND_RNN_BIAS,
        "FRACTAL_ZN_RNN": DD.FORMAT_FRACTAL_ZN_RNN,
        "YUV": DD.FORMAT_YUV,
        "YUV_A": DD.FORMAT_YUV_A,
        "MAX": DD.FORMAT_MAX,
    }

    DTYPE_KEY = 'dtype'
    STRUCT_FORMAT_KEY = 'struct_format'
    DATA_TYPE_TO_DTYPE_MAP = {
        DD.DT_FLOAT: {DTYPE_KEY: np.float32, STRUCT_FORMAT_KEY: 'f'},
        DD.DT_FLOAT16: {DTYPE_KEY: np.float16, STRUCT_FORMAT_KEY: 'e'},
        DD.DT_DOUBLE: {DTYPE_KEY: np.float64, STRUCT_FORMAT_KEY: 'd'},
        DD.DT_INT8: {DTYPE_KEY: np.int8, STRUCT_FORMAT_KEY: 'b'},
        DD.DT_INT16: {DTYPE_KEY: np.int16, STRUCT_FORMAT_KEY: 'h'},
        DD.DT_INT32: {DTYPE_KEY: np.int32, STRUCT_FORMAT_KEY: 'i'},
        DD.DT_INT64: {DTYPE_KEY: np.int64, STRUCT_FORMAT_KEY: 'q'},
        DD.DT_UINT8: {DTYPE_KEY: np.uint8, STRUCT_FORMAT_KEY: 'B'},
        DD.DT_UINT16: {DTYPE_KEY: np.uint16, STRUCT_FORMAT_KEY: 'H'},
        DD.DT_UINT32: {DTYPE_KEY: np.uint32, STRUCT_FORMAT_KEY: 'I'},
        DD.DT_UINT64: {DTYPE_KEY: np.uint64, STRUCT_FORMAT_KEY: 'Q'},
        DD.DT_BOOL: {DTYPE_KEY: np.bool_, STRUCT_FORMAT_KEY: '?'},
        DD.DT_COMPLEX64: {DTYPE_KEY: np.complex64, STRUCT_FORMAT_KEY: '?'},
        DD.DT_COMPLEX128: {DTYPE_KEY: np.complex128, STRUCT_FORMAT_KEY: '?'},
        DD.DT_BF16: {DTYPE_KEY: 'bfloat16', STRUCT_FORMAT_KEY: 'e'},
        DD.DT_UINT1: {DTYPE_KEY: np.uint8, STRUCT_FORMAT_KEY: '?'},
        DD.DT_UNDEFINED: {DTYPE_KEY: np.int8, STRUCT_FORMAT_KEY: 'b'},
        DD.DT_RESOURCE: {DTYPE_KEY: np.int64, STRUCT_FORMAT_KEY: 'q'},
        DD.DT_FLOAT8_E4M3FN: {DTYPE_KEY: "float8_e4m3fn", STRUCT_FORMAT_KEY: 'b'},
        DD.DT_FLOAT8_E5M2: {DTYPE_KEY: "float8_e5m2", STRUCT_FORMAT_KEY: 'b'},
        DD.DT_HIFLOAT8: {DTYPE_KEY: "hifloat8", STRUCT_FORMAT_KEY: 'b'},
    }
    DATA_TYPE_TO_STR_DTYPE_MAP = {
        DD.DT_FLOAT: "float32",
        DD.DT_FLOAT16: "float16",
        DD.DT_DOUBLE: "float64",
        DD.DT_INT8: "int8",
        DD.DT_INT16: "int16",
        DD.DT_INT32: "int32",
        DD.DT_INT64: "int64",
        DD.DT_UINT8: "uint8",
        DD.DT_UINT16: "uint16",
        DD.DT_UINT32: "uint32",
        DD.DT_UINT64: "uint64",
        DD.DT_BOOL: "bool",
        DD.DT_COMPLEX64: "complex64",
        DD.DT_COMPLEX128: "complex128",
        DD.DT_BF16: "bfloat16",
        DD.DT_UINT1: "uint1",
        DD.DT_UNDEFINED: "int8",
        DD.DT_RESOURCE: "int64",
        DD.DT_FLOAT8_E4M3FN: "float8_e4m3fn",
        DD.DT_HIFLOAT8: "hifloat8",
        DD.DT_FLOAT8_E5M2: "float8_e5m2",
    }

    UNPACK_DTYPE = [
        DD.DT_UINT1,
    ]
    CAST_FP32_DTYPE = [
        DD.DT_BF16,
    ]
    FP_LOW_DATA_DTYPE = [
        DD.DT_FLOAT8_E4M3FN,
        DD.DT_HIFLOAT8,
        DD.DT_FLOAT8_E5M2,
    ]

    # Standard
    STANDARD_SUFFIX = ".pb"
    STANDARD_FILE_NAME = 'op_name.output_index.timestamp.pb'

    # Qunat
    QUANT_SUFFIX = ".quant"
    QUANT_FILE_NAME = 'op_name.output_index.timestamp.quant'

    # Offline
    OFFLINE_FILE_NAME = 'op_type.op_name.task_id(.stream_id).timestamp'
    OFFLINE_FILE_NAME_COUNT = 5

    # Standard
    NUMPY_SUFFIX = ".npy"
    NUMPY_FILE_NAME = 'op_name.output_index.timestamp.npy'

    QUANT_OP_NANE_SUFFIX_LIST = [
        "_quant_layer", "_anti_quant_layer", "AscendQuant", "AscendWeightQuant",
        "AntiQuant", "_quant", "_weight_quant", "_anti_quant"
    ]

    DEQUANT_OP_NANE_SUFFIX_LIST = ["_dequant_layer", "AscendDequant", "_dequant"]

    CHAR_FMT = 's'
    INT32_FMT = 'i'
    UINT32_FMT = 'I'
    INT64_FMT = 'q'
    UINT64_FMT = 'Q'

    CHAR_SIZE = 1
    INT16_SIZE = 2
    UINT16_SIZE = 2
    INT32_SIZE = 4
    INT64_SIZE = 8
    UINT32_SIZE = 4
    UINT64_SIZE = 8

    UNPACK_FORMAT = {
        'CHAR': {'FMT': CHAR_FMT, 'SIZE': CHAR_SIZE},
        'UINT32': {'FMT': UINT32_FMT, 'SIZE': UINT32_SIZE},
        'UINT64': {'FMT': UINT64_FMT, 'SIZE': UINT64_SIZE},
        'INT32': {'FMT': INT32_FMT, 'SIZE': INT32_SIZE},
        'INT64': {'FMT': INT64_FMT, 'SIZE': INT64_SIZE},
    }
    ONE_GB = 1 * 1024 * 1024 * 1024
    ONE_MB = 1 * 1024 * 1024
    ONE_HUNDRED_MB = 100 * 1024 * 1024
    DHA_ATOMIC_ADD_INFO_SIZE = 128
    L2_ATOMIC_ADD_INFO_SIZE = 128
    AI_CORE_INFO_SIZE = 256
    DHA_ATOMIC_ADD_STATUS_SIZE = 256
    L2_ATOMIC_ADD_STATUS_SIZE = 256
    AI_CORE_STATUS_SIZE = 1024
    OVERFLOW_CHECK_SIZE = \
        DHA_ATOMIC_ADD_INFO_SIZE + L2_ATOMIC_ADD_INFO_SIZE + AI_CORE_INFO_SIZE + DHA_ATOMIC_ADD_STATUS_SIZE \
        + L2_ATOMIC_ADD_STATUS_SIZE + AI_CORE_STATUS_SIZE
    ACC_TYPE = {
        0: "AIC",
        1: "AIV",
        2: "AICPU",
        3: "SDMA"
    }
    OVERFLOW_DEBUG = ('magic', 'version', 'acc_list')
    ACC_DEBUG = ('valid', 'acc_type', 'rsv', 'data_len', 'data')
    AIC_AIV_DEBUG = (
        'model_id', 'stream_id', 'task_id', 'task_type', 'context_id',
        'thread_id', 'pc_start', 'para_base', 'core_id', 'block_id', 'status'
    )
    SDMA_DEBUG = (
        'model_id', 'stream_id', 'task_id', 'task_type', 'context_id',
        'thread_id', 'src_addr', 'dst_addr', 'channel_id', 'status'
    )
    AICPU_DEBUG = (
        'model_id', 'stream_id', 'task_id', 'task_type', 'context_id', 'cpu_id', 'thread_id', 'status'
    )
    DEBUG_INFO_MAP = {
        "AIC": AIC_AIV_DEBUG,
        "AIV": AIC_AIV_DEBUG,
        "SDMA": SDMA_DEBUG,
        "AICPU": AICPU_DEBUG,
    }
    HEX_FORMAT_ITEM = ("pc_start", "para_base", "src_addr", "dst_addr")

    BUFFER_TYPE_MAP = {DD.L1: 'l1'}
    CONVERT_FAILED_FILE_LIST_NAME = "convert_failed_file_list.txt"
    MAPPING_FILE_NAME = "mapping.csv"
    END_FLAG = "\0"

    AICORE = "AICORE"
    AICPU = "AICPU"
    DEBUG = "DEBUG"
    HCCL = "HCCL"
    FFTSPLUS = "FFTSPLUS"

    TASK_TYPE_MAP = {
        AICORE: '0',
        AICPU: '1',
        DEBUG: '2',
        HCCL: '3',
        FFTSPLUS: '4'
    }
    # task mode
    NORMAL_MODE = 0
    AUTOMATIC_MODE = 1
    MANUAL_MODE = 2
    SPEC_MODE = 3

    FFTS_TIMESTAMP = 'timestamp'

    INVALID_SORT_MODE = 1
    INVALID_THREAD_ID = 2
    INVALID_SLICE_X = 3


    OLD_FILE_FIELD_NUM = 4
    NEW_FILE_FIELD_NUM = 9
    # FFTS/FFTS+ MODE Field
    FFTS_MANUAL_MODE_FIELD = "_lxslice"
    SGT_FIELD = "_sgt_graph"
    LXSLICE_FILED = "lxslice"

    # walk limit
    MAX_WALK_FILE_NUM = 1000
    MAX_WALK_DIR_DEEP_NUM = 50

    SPECIAL_CHAR = ["\n", "\r", "\u007F", "\b", "\f", "\t", "\u000B", "%08", "%0a", "%0b", "%0c", "%0d", "%7f"]

    @property
    def max_top_n(self: any) -> int:
        """
        max top n
        """
        return 10000

    @property
    def min_top_n(self: any) -> int:
        """
        mix top n
        """
        return 1



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import inspect
from typing import Callable
from functools import wraps


class FuncWrapper:
    def __init__(self, check_rules: dict) -> None:
        self.check_rules = check_rules
        self.args_names = []
        self.args_name_set = set()
        self.var_keyword_rule = None

    def is_function_param_valid(self, to_raise, *args, **kwargs) -> bool:
        if kwargs is not None:
            for args_name, arg_value in kwargs.items():
                check_rule = self.check_rules.get(args_name)
                if check_rule is None and args_name not in self.args_name_set and self.var_keyword_rule is not None:
                    check_rule = self.var_keyword_rule
                if check_rule is None:
                    continue

                is_pass = check_rule.check(arg_value, will_raise=to_raise)
                if not is_pass:
                    return bool(is_pass), f"{args_name} is invalid. {str(is_pass)}"

        if args is not None and len(self.args_names) > 0:
            for index, arg_value in enumerate(args):

                if len(self.args_names) > index:
                    args_name = self.args_names[index]
                else:
                    args_name = self.args_names[-1]

                check_rule = self.check_rules.get(args_name)
                if check_rule is None:
                    continue

                is_pass = check_rule.check(arg_value, will_raise=to_raise)
                if not is_pass:
                    return bool(is_pass), f"{args_name} is invalid. {str(is_pass)}"
        return True

    def parse_function(self, func):
        parameters = inspect.signature(func).parameters

        for param in parameters.values():
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                self.var_keyword_rule = self.check_rules.get(param.name)
            else:
                self.args_names.append(param.name)
                if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
                    self.args_name_set.add(param.name)

    def create_wrapper(self, ret_value, to_raise, logger=None) -> Callable:
        def decorator(func) -> Callable:
            self.parse_function(func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                is_pass = self.is_function_param_valid(to_raise, *args, **kwargs)
                if isinstance(is_pass, bool) and is_pass:
                    return func(*args, **kwargs)
                else:
                    if logger is not None and hasattr(logger, "error"):
                        logger.error(str(is_pass))
                    return ret_value

            return wrapper

        return decorator

    def to_return(self, ret_value, logger=None) -> Callable:
        return self.create_wrapper(ret_value, False, logger)

    def to_raise(self) -> Callable:
        return self.create_wrapper(None, True, None)


def validate_params(**check_rules) -> FuncWrapper:
    return FuncWrapper(check_rules)


# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves the advisor result function.
"""
import os

from msprobe.msaccucmp.advisor.advisor_const import AdvisorConst
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils import path_check
from msprobe.msaccucmp.cmp_utils import log


class AdvisorResult:
    """
    Class for generate advisor result
    """

    def __init__(self, match_advisor=False, advisor_type="NA", operator_index="NA", advisor_message="NA"):
        self.match_advisor = match_advisor
        self.advisor_type = advisor_type
        self.operator_index = operator_index
        self.advisor_message = advisor_message

    @staticmethod
    def gen_summary_file(out_path, message_list):
        """
        Generate advisor summary file
        :param out_path: advisor summary file out path
        :param  message_list: summary message
        """
        result_file = os.path.join(out_path, "advisor_summary.txt")
        try:
            path_check.check_write_path_secure(result_file)
            with os.fdopen(os.open(result_file, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                           'w+') as output_file:
                output_file.truncate(0)
                message_list = [message + AdvisorConst.NEW_LINE for message in message_list]
                output_file.writelines(message_list)
        except IOError as io_error:
            log.print_error_log("Failed to save the advisor summary, the reason is %s." % io_error)
        else:
            log.print_info_log('The advisor summary (.txt) is saved in: "%r" .' % result_file)

    def print_advisor_log(self):
        """
        Log and print advisor summary
        """
        log.print_info_log("The summary of the expert advice is as follows: ")
        message_list = [
            AdvisorConst.DETECTION_TYPE + AdvisorConst.COLON + self.advisor_type,
            AdvisorConst.OPERATOR_INDEX + AdvisorConst.COLON + self.operator_index,
            AdvisorConst.ADVISOR_SUGGEST + AdvisorConst.COLON + self.advisor_message
        ]
        for message in message_list:
            log.print_info_log(message)
        return message_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import time
import os
import math

import torch

from msprobe.core.common.utils import CompareException
from msprobe.core.common.file_utils import load_yaml
from msprobe.core.common.const import Const
from msprobe.pytorch.common.log import logger


current_time = time.strftime("%Y%m%d%H%M%S")
API_PRECISION_COMPARE_RESULT_FILE_NAME = "api_precision_compare_result_" + current_time + ".csv"
API_PRECISION_COMPARE_DETAILS_FILE_NAME = "api_precision_compare_details_" + current_time + ".csv"
BENCHMARK_COMPARE_SUPPORT_LIST = ['torch.float16', 'torch.bfloat16', 'torch.float32', "torch.float8_e4m3fn", 
                                  "torch.float8_e5m2"]
API_PRECISION_COMPARE_UNSUPPORT_LIST = ['torch.float64', 'torch.complex64', 'torch.complex128']
ULP_COMPARE_SUPPORT_LIST = ['torch.float16', 'torch.bfloat16', 'torch.float32']
BINARY_COMPARE_UNSUPPORT_LIST = BENCHMARK_COMPARE_SUPPORT_LIST + API_PRECISION_COMPARE_UNSUPPORT_LIST


cur_path = os.path.dirname(os.path.realpath(__file__))
standard_yaml_path = os.path.join(cur_path, "api_precision_standard.yaml")
apis = load_yaml(standard_yaml_path)
absolute_standard_api = apis.get('AbsoluteThreshStandard')
binary_standard_api = apis.get('BinaryCompareStandard')
ulp_standard_api = apis.get('ULPStandard')
thousandth_standard_api = apis.get('ThousandthStandard')
accumulative_error_standard_api = apis.get('AccumulativeErrorStandard')


DETAIL_TEST_ROWS = [
            [
            "API Name", "Bench Dtype", "DEVICE Dtype", "Shape",
            "余弦相似度",
            "最大绝对误差",
            "双百指标",
            "双千指标",
            "双万指标",
            "二进制一致错误率",
            "误差均衡性",
            "均方根误差",
            "小值域错误占比",
            "相对误差最大值",
            "相对误差平均值",
            "inf/nan错误率",
            "相对误差错误率",
            "绝对误差错误率",
            "ULP误差最大值",
            "ULP误差平均值",
            "ULP误差大于阈值占比",
            "Status",
            "Message"
            ]
        ]


precision_configs = {
    torch.float16: {
        'small_value': [
            1e-3
        ],
        'small_value_atol': [
            1e-5
        ]
    },
    torch.bfloat16: {
        'small_value': [
            1e-3
        ],
        'small_value_atol': [
            1e-5
        ]
    },
    torch.float32: {
        'small_value': [
            1e-6
        ],
        'small_value_atol': [
            1e-9
        ]
    }
}


ULP_PARAMETERS = {
    torch.float16: {
        'min_eb': [
            -14
        ],
        'exponent_num': [
            10
        ]
    },
    torch.bfloat16: {
        'min_eb': [
            -126
        ],
        'exponent_num': [
            7
        ]
    },
    torch.float32: {
        'min_eb': [
            -126
        ],
        'exponent_num': [
            23
        ]
    }
}


class ApiPrecisionCompareColumn:
    API_NAME = 'API Name'
    DEVICE_DTYPE = 'DEVICE Dtype'
    SHAPE = 'Shape'
    SMALL_VALUE_ERROR_RATE = '小值域错误占比'
    RMSE = '均方根误差'
    MAX_REL_ERR = '相对误差最大值'
    MEAN_REL_ERR = '相对误差平均值'
    EB = '误差均衡性'
    SMALL_VALUE_ERROR_RATIO = '小值域错误比值'
    SMALL_VALUE_ERROR_STATUS = '小值域判定结果'
    RMSE_RATIO = '均方根误差比值'
    RMSE_STATUS = '均方根误差判定结果'
    MAX_REL_ERR_RATIO = '相对误差最大值比值'
    MAX_REL_ERR_STATUS = '相对误差最大值判定结果'
    MEAN_REL_ERR_RATIO = '相对误差平均值比值'
    MEAN_REL_ERR_STATUS = '相对误差平均值判定结果'
    EB_RATIO = '误差均衡性比值'
    EB_STATUS = '误差均衡性判定结果'
    ERROR_RATE = '二进制一致错误率'
    ERROR_RATE_STATUS = '二进制一致错误率判定结果'
    INF_NAN_ERROR_RATIO = 'inf/nan错误率'
    INF_NAN_ERROR_RATIO_STATUS = 'inf/nan判定结果'
    REL_ERR_RATIO = '相对误差错误率'
    REL_ERR_RATIO_STATUS = '相对误差判定结果'
    ABS_ERR_RATIO = '绝对误差错误率'
    ABS_ERR_RATIO_STATUS = '绝对误差判定结果'
    MEAN_ULP_ERR = 'ULP误差平均值'
    ULP_ERR_PROPORTION = 'ULP误差大于阈值占比'
    ULP_ERR_PROPORTION_RATIO = 'ULP误差大于阈值占比比值'
    ULP_ERR_STATUS = 'ULP误差判定结果'
    REL_ERR_THOUSANDTH = '双千指标'
    REL_ERR_THOUSANDTH_STATUS = '双千指标判定结果'
    FINAL_RESULT = '比对结果'
    ALGORITHM = '比对算法'
    FORWWARD_STATUS = 'Forward Test Success'
    BACKWARD_STATUS = 'Backward Test Success'
    MESSAGE = 'Message'
    
    @staticmethod
    def to_required_columns():
        return [ApiPrecisionCompareColumn.API_NAME, ApiPrecisionCompareColumn.DEVICE_DTYPE, 
                ApiPrecisionCompareColumn.SMALL_VALUE_ERROR_RATE, ApiPrecisionCompareColumn.RMSE, 
                ApiPrecisionCompareColumn.MAX_REL_ERR, ApiPrecisionCompareColumn.MEAN_REL_ERR,
                ApiPrecisionCompareColumn.EB, ApiPrecisionCompareColumn.ERROR_RATE, 
                ApiPrecisionCompareColumn.INF_NAN_ERROR_RATIO, ApiPrecisionCompareColumn.REL_ERR_RATIO, 
                ApiPrecisionCompareColumn.ABS_ERR_RATIO, ApiPrecisionCompareColumn.MEAN_ULP_ERR, 
                ApiPrecisionCompareColumn.ULP_ERR_PROPORTION, ApiPrecisionCompareColumn.REL_ERR_THOUSANDTH]

    @staticmethod
    def get_detail_csv_title():
        return [ApiPrecisionCompareColumn.API_NAME, 
                ApiPrecisionCompareColumn.SMALL_VALUE_ERROR_RATIO, ApiPrecisionCompareColumn.SMALL_VALUE_ERROR_STATUS, 
                ApiPrecisionCompareColumn.RMSE_RATIO, ApiPrecisionCompareColumn.RMSE_STATUS, 
                ApiPrecisionCompareColumn.MAX_REL_ERR_RATIO, ApiPrecisionCompareColumn.MAX_REL_ERR_STATUS, 
                ApiPrecisionCompareColumn.MEAN_REL_ERR_RATIO, ApiPrecisionCompareColumn.MEAN_REL_ERR_STATUS, 
                ApiPrecisionCompareColumn.EB_RATIO, ApiPrecisionCompareColumn.EB_STATUS, 
                ApiPrecisionCompareColumn.INF_NAN_ERROR_RATIO, ApiPrecisionCompareColumn.INF_NAN_ERROR_RATIO_STATUS, 
                ApiPrecisionCompareColumn.REL_ERR_RATIO, ApiPrecisionCompareColumn.REL_ERR_RATIO_STATUS, 
                ApiPrecisionCompareColumn.ABS_ERR_RATIO, ApiPrecisionCompareColumn.ABS_ERR_RATIO_STATUS, 
                ApiPrecisionCompareColumn.ERROR_RATE, ApiPrecisionCompareColumn.ERROR_RATE_STATUS, 
                ApiPrecisionCompareColumn.MEAN_ULP_ERR, ApiPrecisionCompareColumn.ULP_ERR_PROPORTION, 
                ApiPrecisionCompareColumn.ULP_ERR_PROPORTION_RATIO, ApiPrecisionCompareColumn.ULP_ERR_STATUS,
                ApiPrecisionCompareColumn.REL_ERR_THOUSANDTH, ApiPrecisionCompareColumn.REL_ERR_THOUSANDTH_STATUS,
                ApiPrecisionCompareColumn.FINAL_RESULT, ApiPrecisionCompareColumn.ALGORITHM, 
                ApiPrecisionCompareColumn.MESSAGE]
    
    @staticmethod
    def get_result_csv_title():
        return [ApiPrecisionCompareColumn.API_NAME, ApiPrecisionCompareColumn.FORWWARD_STATUS, 
                ApiPrecisionCompareColumn.BACKWARD_STATUS, ApiPrecisionCompareColumn.MESSAGE]


CompareMessage = {
    "topk": "在npu上，topk的入参sorted=False时不生效，会返回有序tensor，而cpu上会返回无序tensor。 如果topk精度不达标，请检查是否是该原因导致的。"
}


def check_dtype_comparable(x, y):
    if x.dtype in Const.FLOAT_TYPE:
        if y.dtype in Const.FLOAT_TYPE:
            return True 
        return False 
    if x.dtype in Const.BOOL_TYPE:
        if y.dtype in Const.BOOL_TYPE:
            return True 
        return False 
    if x.dtype in Const.INT_TYPE:
        if y.dtype in Const.INT_TYPE:
            return True 
        return False
    logger.warning(f"Compare: Unexpected dtype {x.dtype}, {y.dtype}")
    return False


def convert_str_to_float(input_data):
    if isinstance(input_data, str) and input_data.strip() == "":
        msg = 'ERROR: Input data is an empty string'
        raise CompareException(CompareException.INVALID_DATA_ERROR, msg)
    try:
        float_data = float(input_data)
        return float_data
    except ValueError as e:
        msg = 'ERROR: Input data cannot be converted to float'
        raise CompareException(CompareException.INVALID_DATA_ERROR, msg) from e


def is_inf_or_nan(x):
    return math.isnan(x) or math.isinf(x)


def handle_infinity(x, y, column_name):
    if math.isinf(x) and math.isinf(y):
        if x == y:
            return float("nan"), True, f"{column_name}同为同号inf或nan\n"
        else:
            return float("nan"), False, f"{column_name}inf或nan不一致\n"
    else:
        return float("nan"), False, f"{column_name}inf或nan不一致\n"


def handle_nan(x, y, column_name):
    if math.isnan(x) and math.isnan(y):
        return float("nan"), True, f"{column_name}同为同号inf或nan\n"
    else:
        return float("nan"), False, f"{column_name}inf或nan不一致\n"


def check_inf_or_nan(x, y, column_name):
    if math.isinf(x) or math.isinf(y):
        return handle_infinity(x, y, column_name)
    else:
        return handle_nan(x, y, column_name)
    


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Union
from msprobe.infer.utils.check.checker import Checker, CheckResult, rule


class ObjectChecker(Checker):
    @rule()
    def is_attr_valid(self, key, check_rule: Checker, default_value=None) -> Union["ObjectChecker", CheckResult]:
        value = getattr(self.instance, key, default_value)
        return check_rule.check(value)
    
    @rule()
    def is_attrs_valid(self, default_value=None, **attr_rules: Checker) -> Union["ObjectChecker", CheckResult]:
        is_pass = self.not_none()
        if not is_pass:
            return is_pass
        
        err_msgs = []
        for attr_name, attr_rule in attr_rules.items():
            attr_value = getattr(self.instance, attr_name, default_value)
            rule_is_pass = attr_rule.check(attr_value)
            if not rule_is_pass:
                is_pass = False 
                err_msgs.append(f"{attr_name} is invalid. {str(rule_is_pass)}")
        
        return is_pass, "\n".join(err_msgs)


# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
SelectMode class.
This class mainly involves functions for selecting operators.
"""

from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_rule import CompareRule
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.vector_cmp.range_manager import range_manager


class SelectMode(range_manager.RangeManager):
    """
    The subclass of range manager
    """

    def __init__(self: any, input_str: str) -> None:
        super(range_manager.RangeManager, self).__init__()
        self.selected_op = self._parse_input_str(input_str)

    @staticmethod
    def _parse_input_str(input_str: str) -> list:
        input_operators = []
        index_list = input_str.split(',')
        for item in index_list:
            value = item.strip()
            if not value:
                continue
            if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, value):
                log.print_error_log('The index (%s) is invalid, just supports '
                                    '"index_1, index_2, ...", the value is zero or a positive number.' % input_str)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            input_operators.append(int(value))
        selected_operators = list(set(input_operators))
        index_list_sorted = sorted(selected_operators)
        return index_list_sorted

    def get_all_ops(self: any, compare_rule: CompareRule) -> list:
        """
        Get all operators according to indexes:
        :param compare_rule: the compare rule
        :return: operator list
        """
        op_list = [op for op in compare_rule.fusion_info.op_list if op.attr.get_op_sequence() in self.selected_op]
        return self._get_op_list(op_list, compare_rule)

    def check_input_valid(self: any, op_count: int) -> None:
        """
        Check range valid:
        :param op_count: the op count
        """
        op_list = []
        out_of_range_list = []
        for value in self.selected_op:
            if value <= op_count:
                op_list.append(value)
            else:
                out_of_range_list.append(value)
        if out_of_range_list:
            invalid_ops = str(out_of_range_list)
            message = 'Index %s out of range ' % invalid_ops + '[1,%d].' % op_count
            if not op_list:
                log.print_error_log(message)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
            log.print_warn_log(message)
        self.selected_op = op_list
        log.print_info_log('Operator list %s is valid.' % (str(self.selected_op)))
